from io import StringIO
from test.testbase import DbIsolatedTest, run_as_sync
from unittest.mock import patch

from db.python.layers import ParticipantLayer
from metamist.parser.generic_parser import ParsedParticipant
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    ParticipantUpsertInternal,
    SampleUpsertInternal,
)
from scripts.parse_ont_sheet import OntParser


class TestOntSampleSheetParser(DbIsolatedTest):
    """Test the TestOntSampleSheetParser"""

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    async def test_simple_sheet(self, mock_graphql_query):
        """
        Test importing a two rows, forms objects and checks response
        - MOCKS:
            - get_participant_id_map_by_external_ids
            - get_sample_id_map_by_external
            - get_sequence_ids_for_sample_ids_by_type
        """

        player = ParticipantLayer(self.connection)
        await player.upsert_participants(
            [
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'Sample01'},
                    samples=[
                        SampleUpsertInternal(
                            external_ids={PRIMARY_EXTERNAL_ORG: 'Sample01'},
                        )
                    ],
                )
            ]
        )

        mock_graphql_query.side_effect = self.run_graphql_query_async

        rows = [
            'Sequencing_date,Experiment name,Sample ID,Protocol,Flow cell,Barcoding,Device,Flowcell ID,MUX total,Basecalling,Fail FASTQ filename,Pass FASTQ filename',
            '10/12/2034,PBXP_Awesome,Sample01,LSK1,PRO002,None,PromethION,XYZ1,7107,4.0.11+f1071ce,Sample01_fail.fastq.gz,Sample01_pass.fastq.gz',
            '21/10/2021,PBXP_Awesome,Sample02,LSK1,PRO002,None,PromethION,XYZ1,8056,4.0.11+f1071ce,Sample02_fail.fastq.gz,Sample02_pass.fastq.gz',
        ]
        parser = OntParser(
            search_locations=[],
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
        )

        parser.skip_checking_gcs_objects = True
        fs = [
            'Sample01_pass.fastq.gz',
            'Sample01_fail.fastq.gz',
            'Sample02_pass.fastq.gz',
            'Sample02_fail.fastq.gz',
        ]
        parser.filename_map = {k: 'gs://BUCKET/FAKE/' + k for k in fs}
        parser.skip_checking_gcs_objects = True

        file_contents = '\n'.join(rows)
        participants: list[ParsedParticipant]
        summary, participants = await parser.parse_manifest(
            StringIO(file_contents), delimiter=',', dry_run=True
        )

        participants_to_add = summary.participants.insert
        participants_to_update = summary.participants.update
        samples_to_add = summary.samples.insert
        samples_to_update = summary.samples.update
        sequencing_to_add = summary.assays.insert
        sequencing_to_update = summary.assays.update

        self.assertEqual(1, participants_to_add)
        self.assertEqual(1, participants_to_update)
        self.assertEqual(1, samples_to_add)
        self.assertEqual(2, sequencing_to_add)
        self.assertEqual(1, samples_to_update)
        self.assertEqual(0, sequencing_to_update)

        meta_dict = {
            'barcoding': 'None',
            'basecalling': '4.0.11+f1071ce',
            'device': 'PromethION',
            'experiment_name': 'PBXP_Awesome',
            'flow_cell': 'PRO002',
            'flowcell_id': 'XYZ1',
            'mux_total': 7107,
            'protocol': 'LSK1',
            'reads': [
                {
                    'basename': 'Sample01_pass.fastq.gz',
                    'checksum': None,
                    'class': 'File',
                    'location': 'gs://BUCKET/FAKE/Sample01_pass.fastq.gz',
                    'size': None,
                    'datetime_added': None,
                }
            ],
            'reads_type': 'fastq',
            'sequencing_date': '10/12/2034',
            'sequencing_platform': 'oxford-nanopore',
            'sequencing_technology': 'long-read',
            'sequencing_type': 'genome',
        }

        seqgroup_meta = {
            'failed_reads': [
                [
                    {
                        'location': 'gs://BUCKET/FAKE/Sample01_fail.fastq.gz',
                        'basename': 'Sample01_fail.fastq.gz',
                        'class': 'File',
                        'checksum': None,
                        'size': None,
                        'datetime_added': None,
                    }
                ]
            ],
        }
        self.maxDiff = None
        sequencing_group = participants[0].samples[0].sequencing_groups[0]
        self.assertDictEqual(seqgroup_meta, sequencing_group.meta)
        self.assertDictEqual(meta_dict, sequencing_group.assays[0].meta)



# noqa: B006 pylint: disable=C0302, E1101

from datetime import datetime
import unittest
import unittest.mock
from io import StringIO
from metamist.audit.adapters import StorageClient
from metamist.audit.cli.upload_bucket_audit import AuditOrchestrator
from metamist.audit.cli.review_audit_results import review_rows
from metamist.audit.cli.delete_from_audit_results import delete_files_from_report
from metamist.audit.data_access import MetamistDataAccess, GCSDataAccess
from metamist.audit.services import (
    AuditAnalyzer,
    FileMatchingService,
    Reporter,
    BucketAuditLogger,
)
from metamist.audit.models import (
    SequencingGroup,
    Assay,
    Analysis,
    AuditConfig,
    FileType,
    AuditResult,
    AuditReportEntry,
    FileMetadata,
)
from test.data.audit_fixtures.audit_test import (
    ANALYSES_WITH_MALFORMED_TIMESTAMPS_GQL_RESPONSE,
    SG_ASSAYS_RESPONSE_WITH_SECONDARY_FILES,
    ENUMS_QUERY_RESULT,
    SEQUENCING_GROUPS,
    SR_ILLUMINA_EXOME_SGS,
    ANALYSES,
    ALL_SG_ASSAYS_RESPONSE,
    SR_ILLUMINA_EXOME_SGS_ASSAYS_RESPONSE,
    ALL_SG_ANALYSES_RESPONSE,
    SR_ILLUMINA_EXOME_SGS_CRAM_RESPONSE,
)
from test.data.audit_fixtures.bucket_files import UPLOAD_BUCKET_FILES, MAIN_BUCKET_FILES
from test.testbase import run_as_sync
from types import SimpleNamespace
from cpg_utils import to_path


class TestAudit(unittest.TestCase):  # pylint: disable=too-many-instance-attributes, too-many-public-methods
    """Test the audit module"""

    def setUp(self):
        """Set up test fixtures."""
        # Start the storage client mock before creating GCSDataAccess
        self.storage_client_patcher = unittest.mock.patch(
            'metamist.audit.adapters.storage_client.storage.Client'
        )
        self.mock_storage_client = self.storage_client_patcher.start()

        # Mock get_bucket_name to avoid needing real config
        self.bucket_patcher = unittest.mock.patch.object(
            GCSDataAccess, 'get_bucket_name'
        )
        self.mock_get_bucket_name = self.bucket_patcher.start()
        self.mock_get_bucket_name.side_effect = [
            'cpg-dataset-main',
            'cpg-dataset-main-upload',
            'cpg-dataset-main-analysis',
        ]

        self.metamist_data_access = MetamistDataAccess()
        self.gcs_data_access = GCSDataAccess('dataset', 'test')
        self.file_matcher = FileMatchingService()
        self.audit_analyzer = AuditAnalyzer()
        self.audit_logs = BucketAuditLogger('dataset', 'test')
        self.reporter = Reporter(self.gcs_data_access, self.audit_logs)

        self.sample_csv_content = """File Path,File Size,SG ID,SG Type,SG Technology,SG Platform,Assay ID,Sample ID,Sample External ID,Participant ID,Participant External ID,CRAM Analysis ID,CRAM Path,Action,Review Comment
gs://cpg-dataset-main-upload/file1.bam,2048000000,SG01,exome,short-read,illumina,1,S01,EXT001,1,"P001",1,cram,delete,File is old and replaced by newer upload
gs://cpg-dataset-main-upload/file2.bam,512000000,SG02,genome,short-read,illumina,2,S02,EXT002,2,"P002",2,cram,review,File not found in main bucket"""

    def tearDown(self):
        """Clean up test fixtures."""
        self.storage_client_patcher.stop()
        self.bucket_patcher.stop()

    def test_file_matching_service(self):
        """Test FileMatchingService"""
        self.assertIsInstance(self.file_matcher, FileMatchingService)
        self.assertIsInstance(self.audit_analyzer.file_matcher, FileMatchingService)

    def test_gcs_data_access_setup(self):
        """
        First test that GCSDataAccess initializes correctly with mocked buckets
        since these are used frequently in the tests.
        """
        # Verify that the bucket names are set correctly from our mocks
        self.assertEqual(self.gcs_data_access.main_bucket, 'cpg-dataset-main')
        self.assertEqual(self.gcs_data_access.upload_bucket, 'cpg-dataset-main-upload')
        self.assertEqual(
            self.gcs_data_access.analysis_bucket, 'cpg-dataset-main-analysis'
        )
        self.assertEqual(self.gcs_data_access.dataset, 'dataset')
        self.assertEqual(self.gcs_data_access.gcp_project, 'test')

    # ===== METAMIST DATA ACCESS TESTS =====
    # These test the data access layer - mapping GraphQL responses to domain models
    @run_as_sync
    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query_async')
    async def test_validate_enums(self, mock_query_async):
        """Test MetamistDataAccess.validate_enums"""
        mock_query_async.return_value = ENUMS_QUERY_RESULT
        config = AuditConfig(
            dataset='dataset',
            sequencing_types=('genome', 'exome'),
            sequencing_technologies=('short-read', 'long-read'),
            sequencing_platforms=('illumina', 'pacbio', 'ont'),
            analysis_types=('cram', 'vcf'),
            file_types=tuple(list(FileType)),
        )
        result = await self.metamist_data_access.validate_metamist_enums(config=config)
        self.assertEqual(config, result)

    @run_as_sync
    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query_async')
    async def test_validate_enums_failure(self, mock_query_async):
        """Test MetamistDataAccess.validate_enums invalid input"""
        mock_query_async.return_value = ENUMS_QUERY_RESULT
        with self.assertRaises(ValueError):
            await self.metamist_data_access.validate_metamist_enums(
                config=AuditConfig(
                    dataset='dataset',
                    sequencing_types=('genome', 'exome'),
                    sequencing_technologies=(
                        'short-read',
                        'long-read',
                        'future-technology',
                    ),
                    sequencing_platforms=('illumina', 'pacbio', 'ont'),
                    analysis_types=('cram', 'vcf'),
                    file_types=tuple(list(FileType)),
                )
            )

    @run_as_sync
    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query_async')
    async def test_get_all_sequencing_groups(self, mock_query_async):
        """Test MetamistDataAccess.get_sequencing_groups"""
        mock_query_async.return_value = ALL_SG_ASSAYS_RESPONSE

        result = await self.metamist_data_access.get_sequencing_groups(
            dataset='dataset',
            sequencing_types=['genome', 'exome'],
            sequencing_technologies=['short-read', 'long_read'],
            sequencing_platforms=['illumina', 'pacbio', 'ont'],
        )
        self.assertEqual(len(result), 7)
        for sg in result:
            self.assertIsInstance(sg, SequencingGroup)
            for assay in sg.assays:
                self.assertIsInstance(assay, Assay)
                for rf in assay.read_files:
                    self.assertIsInstance(rf, FileMetadata)

    @run_as_sync
    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query_async')
    async def test_get_sequencing_groups_with_secondary_files(self, mock_query_async):
        """Test MetamistDataAccess.get_sequencing_groups"""
        mock_query_async.return_value = SG_ASSAYS_RESPONSE_WITH_SECONDARY_FILES

        result = await self.metamist_data_access.get_sequencing_groups(
            dataset='dataset',
            sequencing_types=['genome', 'exome'],
            sequencing_technologies=['short-read', 'long_read'],
            sequencing_platforms=['illumina', 'pacbio', 'ont'],
        )
        self.assertEqual(len(result[0].assays[0].read_files), 2)

    @run_as_sync
    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query_async')
    async def test_get_filtered_sequencing_groups(self, mock_query_async):
        """Test MetamistDataAccess.get_sequencing_groups filtered"""
        mock_query_async.return_value = SR_ILLUMINA_EXOME_SGS_ASSAYS_RESPONSE

        result = await self.metamist_data_access.get_sequencing_groups(
            dataset='dataset',
            sequencing_types=['exome'],
            sequencing_technologies=['short-read'],
            sequencing_platforms=['illumina'],
        )

        self.assertEqual(len(result), 1)
        sg = result[0]
        self.assertIsInstance(sg, SequencingGroup)
        self.assertEqual(sg.id, 'SG01_2')
        self.assertEqual(sg.technology, 'short-read')
        self.assertEqual(sg.platform, 'illumina')
        self.assertEqual(sg.sample.external_id, 'EXT001')
        self.assertEqual(sg.sample.participant.external_id, 'P001')

    @run_as_sync
    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query')
    async def test_get_dataset_cohort(self, mock_query):
        """Test MetamistDataAccess.get_dataset_cohort"""
        mock_query.return_value = {
            'project': {'cohorts': [{'id': 'C01', 'name': 'Cohort 1'}]}
        }
        cohort_id = self.metamist_data_access.graphql_client.get_dataset_cohort(
            'dataset', 'Cohort 1'
        )
        self.assertEqual(cohort_id, 'C01')

    @run_as_sync
    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query')
    async def test_get_dataset_cohort_create_new(self, mock_query):
        """Test MetamistDataAccess.get_dataset_cohort where none are found so we create new"""
        mock_query.side_effect = [
            {  # Query 1 - for the existing cohorts
                'project': {'cohorts': []}
            },
            {  # Query 2 - a mutation query to create a new cohort
                'cohort': {
                    'createCohortFromCriteria': {'id': 'C01', 'name': 'Cohort 1'}
                }
            },
        ]
        cohort_id = self.metamist_data_access.graphql_client.get_dataset_cohort(
            'dataset', 'Cohort 1'
        )
        self.assertEqual(cohort_id, 'C01')

    @run_as_sync
    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query')
    async def test_get_audit_deletion_analysis(self, mock_query):
        """Test MetamistDataAccess.get_audit_deletion_analyses"""
        mock_query.return_value = {
            'project': {
                'analyses': [
                    {
                        'id': 1,
                        'timestampCompleted': '2025-09-01T16:33:00',
                        'output': 'gs://cpg-dataset-main-analysis/audit_results/20250901_1633/deleted_file.csv',
                        'meta': {},
                    },
                    {
                        'id': 2,
                        'timestampCompleted': '2025-09-01T16:43:00',
                        'output': 'gs://cpg-dataset-main-analysis/audit_results/20250901_1643/deleted_file.csv',
                        'meta': {},
                    },
                ],
            },
        }

        result = self.metamist_data_access.get_audit_deletion_analysis(
            dataset='dataset',
            output='gs://cpg-dataset-main-analysis/audit_results/20250901_1633/deleted_file.csv',
        )

        self.assertIsInstance(result, dict)
        self.assertEqual(result['id'], 1)

    @run_as_sync
    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query')
    async def test_get_audit_deletion_analysis_none(self, mock_query):
        """Test MetamistDataAccess.get_audit_deletion_analyses"""
        mock_query.return_value = {
            'project': {
                'analyses': [],
            },
        }

        result = self.metamist_data_access.get_audit_deletion_analysis(
            dataset='dataset',
            output='gs://cpg-dataset-main-analysis/audit_results/20250901_1633/deleted_file.csv',
        )

        self.assertIsNone(result)

    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query')
    def test_create_audit_deletion_analysis(self, mock_query):
        """Test MetamistDataAccess.create_audit_deletion_analysis"""
        mock_query.side_effect = [
            {'project': {'cohorts': [{'id': 'C01', 'name': 'Cohort 1'}]}},
            {'analysis': {'createAnalysis': {'id': 1}}},
        ]
        result = self.metamist_data_access.create_audit_deletion_analysis(
            dataset='dataset',
            cohort_name='Cohort 1',
            audited_report_name='files_to_delete',
            deletion_report_path='gs://cpg-dataset-main-analysis/audit_results/20250901_1633/files_to_delete.csv',
            stats={},
        )

        self.assertEqual(result, 1)

    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query')
    def test_update_audit_deletion_analysis(self, mock_query):
        """Test MetamistDataAccess.update_audit_deletion_analysis"""
        report_path = 'gs://cpg-dataset-main-analysis/audit_results/20250901_1633/files_to_delete.csv'
        mock_query.side_effect = [
            {'project': {'analyses': [{'id': 1, 'output': report_path, 'meta': {}}]}},
            {'analysis': {'updateAnalysis': {'id': 1}}},
        ]
        existing_audit_deletion = self.metamist_data_access.get_audit_deletion_analysis(
            dataset='dataset',
            output=report_path,
        )
        self.assertEqual(existing_audit_deletion['output'], report_path)
        result = self.metamist_data_access.update_audit_deletion_analysis(
            existing_analysis=existing_audit_deletion,
            audited_report_name='files_to_delete',
            stats={},
        )

        self.assertEqual(result, 1)

    @run_as_sync
    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query_async')
    async def test_get_analyses_for_sequencing_groups(self, mock_query_async):
        """Test MetamistDataAccess.get_analyses_for_sequencing_groups"""
        mock_query_async.return_value = ALL_SG_ANALYSES_RESPONSE

        result = await self.metamist_data_access.get_analyses_for_sequencing_groups(
            dataset='dataset',
            sg_ids=[sg.id for sg in SEQUENCING_GROUPS.values()],
            analysis_types=['cram', 'vcf'],
        )

        self.assertEqual(len(result), 7)
        for analysis in result:
            self.assertIsInstance(analysis, Analysis)
            self.assertIn(analysis.id, [1, 2, 3, 4, 5, 6, 7])
            self.assertIn(analysis.type, ['cram', 'vcf'])

    @run_as_sync
    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query_async')
    async def test_get_analyses_alternate(self, mock_query_async):
        """Test MetamistDataAccess.get_analyses_for_sequencing_groups"""
        mock_query_async.return_value = ANALYSES_WITH_MALFORMED_TIMESTAMPS_GQL_RESPONSE

        result = await self.metamist_data_access.get_analyses_for_sequencing_groups(
            dataset='dataset',
            sg_ids=['SG101'],
            analysis_types=['cram'],
        )

        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[0], Analysis)
        self.assertEqual(result[0].id, 101)
        self.assertEqual(result[0].timestamp_completed, datetime.min)
        self.assertEqual(
            result[0].output_path, 'gs://cpg-dataset-main-upload/2023-05-01/EXT101.cram'
        )

    @run_as_sync
    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query_async')
    async def test_get_filtered_analyses_for_sequencing_groups(self, mock_query_async):
        """Test MetamistDataAccess.get_analyses_for_sequencing_groups filtered"""
        mock_query_async.return_value = SR_ILLUMINA_EXOME_SGS_CRAM_RESPONSE

        result = await self.metamist_data_access.get_analyses_for_sequencing_groups(
            dataset='dataset',
            sg_ids=[sg.id for sg in SR_ILLUMINA_EXOME_SGS],
            analysis_types=['cram'],
        )

        self.assertEqual(len(result), 1)
        analysis = result[0]
        self.assertIsInstance(analysis, Analysis)
        self.assertEqual(analysis.id, 2)
        self.assertEqual(analysis.sequencing_group_id, 'SG01_2')
        self.assertEqual(analysis.type, 'cram')
        self.assertTrue(analysis.is_cram)
        self.assertEqual(
            analysis.output_path, 'gs://cpg-dataset-main/exome/cram/SG01_2.cram'
        )

    # ===== GCS DATA ACCESS TESTS =====
    # These test the data access layer - mapping GCS responses to domain models

    @unittest.mock.patch.object(GCSDataAccess, 'list_files_in_bucket')
    def test_list_read_files_in_bucket(self, mock_list_files):
        """Test listing the files in a bucket"""
        mock_list_files.return_value = [
            f
            for f in UPLOAD_BUCKET_FILES
            if str(f.filepath).endswith(FileType.all_read_extensions())
        ]

        blobs = self.gcs_data_access.list_files_in_bucket(
            bucket_name='cpg-dataset-main-upload',
            file_types=[FileType.FASTQ, FileType.BAM, FileType.CRAM],
        )

        self.assertEqual(len(blobs), 10)

    @unittest.mock.patch.object(GCSDataAccess, 'list_files_in_bucket')
    def test_list_all_files_in_bucket(self, mock_list_files):
        """Test listing the files in a bucket"""
        mock_list_files.return_value = [
            f
            for f in UPLOAD_BUCKET_FILES
            if str(f.filepath).endswith(FileType.all_extensions())
        ]

        blobs = self.gcs_data_access.list_files_in_bucket(
            bucket_name='cpg-dataset-main-upload',
            file_types=list(FileType),
        )

        self.assertEqual(len(blobs), 11)

    @unittest.mock.patch.object(StorageClient, 'check_blobs')
    def test_validate_cram_files(self, mock_check_blobs):
        """Test validating CRAM files"""
        mock_check_blobs.return_value = [f.filepath for f in MAIN_BUCKET_FILES[:2]]

        found, missing = self.gcs_data_access.validate_cram_files(
            cram_paths=[f.filepath for f in MAIN_BUCKET_FILES[:2]]
        )

        self.assertEqual(len(found), 2)
        self.assertEqual(len(missing), 0)
        self.assertEqual(found[0], to_path('gs://cpg-dataset-main/cram/SG01_1.cram'))
        self.assertEqual(
            found[1], to_path('gs://cpg-dataset-main/exome/cram/SG01_2.cram')
        )

    @unittest.mock.patch.object(StorageClient, 'check_blobs')
    def test_validate_cram_files_with_missing(self, mock_check_blobs):
        """Test validating CRAM files where one is missing"""
        # Mock check_blobs to return only 2 of the 3 requested files
        mock_check_blobs.return_value = [f.filepath for f in MAIN_BUCKET_FILES[:2]]

        found, missing = self.gcs_data_access.validate_cram_files(
            cram_paths=[f.filepath for f in MAIN_BUCKET_FILES[:3]]
        )

        self.assertEqual(len(found), 2)
        self.assertEqual(len(missing), 1)
        self.assertEqual(found[0], to_path('gs://cpg-dataset-main/cram/SG01_1.cram'))
        self.assertEqual(
            found[1], to_path('gs://cpg-dataset-main/exome/cram/SG01_2.cram')
        )
        self.assertEqual(missing[0], to_path('gs://cpg-dataset-main/cram/SG02.cram'))

    # ===== AUDIT ANALYZER TESTS =====
    # These test the core business logic - most complex tests

    def test_analyze_sequencing_groups_basic(self):
        """Test basic audit analysis"""
        # Test the core analysis
        result = self.audit_analyzer.analyze_sequencing_groups(
            list(SEQUENCING_GROUPS.values()),
            UPLOAD_BUCKET_FILES,
            list(ANALYSES.values()),
        )
        self.assertIsInstance(result, AuditResult)

        # Should have files to delete
        self.assertGreater(len(result.files_to_delete), 0)
        # Should have files to review (uningested file)
        self.assertGreater(len(result.files_to_review), 0)
        # Should mark SG as complete
        self.assertEqual(len(result.unaligned_sequencing_groups), 1)
        self.assertEqual(result.unaligned_sequencing_groups[0].id, 'SG04')

    def test_analyze_with_moved_files(self):
        """Test moved file detection"""
        analyses: list[Analysis] = []
        result = self.audit_analyzer.analyze_sequencing_groups(
            list(SEQUENCING_GROUPS.values()), UPLOAD_BUCKET_FILES, analyses
        )

        # Should detect the moved files
        self.assertGreater(len(result.moved_files), 0)
        self.assertEqual(
            result.moved_files[0].filepath,
            'gs://cpg-dataset-main-upload/unknown_files/unknown_R1.fq',
        )
        self.assertEqual(
            result.moved_files[1].filepath,
            'gs://cpg-dataset-main-upload/unknown_files/unknown_R2.fq',
        )

    def test_analyze_with_uningested_files(self):
        """Test uningested file detection"""
        analyses: list[Analysis] = []
        result = self.audit_analyzer.analyze_sequencing_groups(
            list(SEQUENCING_GROUPS.values()), UPLOAD_BUCKET_FILES, analyses
        )

        self.assertGreater(len(result.files_to_review), 0)
        for entry in result.files_to_review:
            self.assertIsNotNone(entry.filepath)

    # ===== FILE MATCHING SERVICE TESTS =====
    # These test the file matching logic
    def test_find_original_analysis_files(self):
        """Test matching Analysis files with bucket files based on checksum"""
        analyses = list(ANALYSES.values())
        self.file_matcher.find_original_analysis_files(analyses, MAIN_BUCKET_FILES)

        for analysis in analyses:
            if analysis.id == 7:
                self.assertIsNotNone(analysis.original_file)

    def test_find_uningested_files(self):
        """Test finding uningested files"""
        metamist_files: list[FileMetadata] = []
        bucket_files = UPLOAD_BUCKET_FILES[:1]

        moved_files = self.file_matcher.find_moved_files(metamist_files, bucket_files)
        self.assertEqual(len(moved_files), 0)

        uningested = self.file_matcher.find_uningested_files(
            metamist_files, bucket_files, moved_files
        )

        self.assertEqual(len(uningested), 1)
        self.assertEqual(
            uningested[0].filepath,
            to_path('gs://cpg-dataset-main-upload/2025-01-01/bams/EXT001.bam'),
        )

    # ===== INTEGRATION TESTS =====
    # These test multiple components working together

    @run_as_sync
    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query_async')
    async def test_audit_config_instantiation_from_inputs(self, mock_query_async):
        """Test configuration creation from inputs"""
        config_args = SimpleNamespace(
            dataset='dataset',
            sequencing_types=('genome', 'exome'),
            sequencing_technologies=('short-read', 'long-read'),
            sequencing_platforms=('illumina', 'pacbio', 'ont'),
            analysis_types=('cram', 'vcf'),
            file_types=tuple(list(FileType)),
            excluded_prefixes=('<EXCLUDED_PREFIX>',),
            results_folder='results_folder',
        )
        mock_query_async.return_value = ENUMS_QUERY_RESULT

        config = AuditConfig.from_cli_args(config_args)
        self.assertIsInstance(config, AuditConfig)

    @run_as_sync
    @unittest.mock.patch('metamist.audit.adapters.graphql_client.query_async')
    @unittest.mock.patch.object(GCSDataAccess, 'list_files_in_bucket')
    async def test_full_audit_workflow_write_reports(
        self, mock_list_files, mock_query_async
    ):
        """Integration test - combines multiple components"""
        mock_query_async.side_effect = [
            # First call: get sequencing groups
            ALL_SG_ASSAYS_RESPONSE,
            # Second call: get analyses for sequencing groups
            ALL_SG_ANALYSES_RESPONSE,
        ]
        mock_list_files.return_value = UPLOAD_BUCKET_FILES

        orchestrator = AuditOrchestrator(
            self.metamist_data_access,
            self.gcs_data_access,
            self.audit_analyzer,
            Reporter(self.gcs_data_access, BucketAuditLogger('dataset', 'test')),
            BucketAuditLogger('dataset', 'test'),
        )

        config = AuditConfig(
            dataset='dataset',
            sequencing_types=('genome', 'exome'),
            sequencing_technologies=('short-read', 'long-read'),
            sequencing_platforms=('illumina', 'pacbio', 'ont'),
            analysis_types=('cram', 'vcf'),
            file_types=tuple(list(FileType)),
            excluded_prefixes=('<EXCLUDED_PREFIX>',),
        )

        # This would normally write files, so mock that too
        with unittest.mock.patch.object(
            Reporter, 'write_audit_reports'
        ) as mock_write_reports:
            await orchestrator.run_audit(config)
            self.assertTrue(mock_write_reports.called)

    # ===== REPORTER SERVICE TESTS =====
    # These test the report writing and reading logic
    def test_audit_report_entry_to_and_from_dict(self):
        """Test AuditReportEntry.from_dict"""
        data = {
            'filepath': 'gs://cpg-dataset-main-upload/file1.bam',
            'filesize': 2048000000,
            'assay_id': 1,
            'sg_id': 'SG01',
            'sg_type': 'exome',
            'sg_tech': 'short-read',
            'sg_platform': 'illumina',
            'sample_id': 'S01',
            'sample_external_ids': 'EXT001',
            'participant_id': 1,
            'participant_external_ids': ['P001'],
            'cram_analysis_id': 1,
            'cram_file_path': 'cram',
            'action': 'delete',
            'review_comment': 'File is old and replaced by newer upload',
        }
        entry = AuditReportEntry.from_report_dict(data)
        self.assertIsInstance(entry, AuditReportEntry)
        self.assertEqual(entry.filepath, data['filepath'])
        self.assertEqual(entry.filesize, data['filesize'])
        self.assertEqual(entry.sg_id, data['sg_id'])
        self.assertEqual(entry.sg_type, data['sg_type'])
        self.assertEqual(entry.sg_tech, data['sg_tech'])
        self.assertEqual(entry.sg_platform, data['sg_platform'])
        self.assertEqual(entry.sample_id, data['sample_id'])
        self.assertEqual(
            entry.participant_external_ids, data['participant_external_ids']
        )
        self.assertEqual(entry.cram_analysis_id, data['cram_analysis_id'])
        self.assertEqual(entry.cram_file_path, data['cram_file_path'])
        self.assertEqual(entry.action, data['action'])
        self.assertEqual(entry.review_comment, data['review_comment'])

        entry_dict = entry.to_report_dict()
        fields_to_headers = {
            v: k for k, v in AuditReportEntry.HEADER_TO_FIELD_MAP.items()
        }
        self.assertEqual(
            entry_dict,
            {fields_to_headers[header]: value for header, value in data.items()},
        )

    def test_get_report_rows_from_name_file_exists(self):
        """Test the Reporter.get_report_rows_from_name method."""
        mock_path = unittest.mock.MagicMock()
        mock_path.exists.return_value = True

        # Mock the file opening to return StringIO with our CSV content
        mock_file = StringIO(self.sample_csv_content)
        mock_path.open.return_value.__enter__.return_value = mock_file

        rows = self.reporter.get_report_rows(mock_path)
        self.assertEqual(len(rows), 2)
        self.assertIsInstance(rows[0], AuditReportEntry)
        self.assertEqual(rows[0].filepath, 'gs://cpg-dataset-main-upload/file1.bam')
        self.assertEqual(
            rows[0].filesize, '2048000000'
        )  # Note: will be string from CSV
        self.assertEqual(rows[0].sg_id, 'SG01')
        self.assertEqual(rows[0].action.upper(), 'DELETE')
        self.assertEqual(rows[1].filepath, 'gs://cpg-dataset-main-upload/file2.bam')
        self.assertEqual(rows[1].action.upper(), 'REVIEW')

    def test_get_report_rows_file_not_exists(self):
        """Test reading CSV when file doesn't exist."""
        mock_path = unittest.mock.MagicMock()
        mock_path.exists.return_value = False

        with unittest.mock.patch.object(self.audit_logs, 'error') as mock_error:
            mock_error.return_value = None  # No-op for logging
            rows = self.reporter.get_report_rows(mock_path)

            self.assertEqual(rows, [])
            self.assertTrue(mock_error.called)

    def test_get_report_stats_from_file(self):
        """Test the Reporter.get_report_stats method."""
        mock_path = unittest.mock.MagicMock()
        mock_path.exists.return_value = True
        mock_file = StringIO(self.sample_csv_content)
        mock_path.open.return_value.__enter__.return_value = mock_file

        stats = self.reporter.get_report_stats(mock_path)

        expected_total_size = 2048000000 + 512000000  # Sum of file sizes
        self.assertEqual(stats['total_size'], expected_total_size)
        self.assertEqual(stats['file_count'], 2)

    def test_get_report_entries_stats(self):
        """Test the Reporter.get_report_entries_stats method."""
        entries = [
            AuditReportEntry(filepath='file1.bam', filesize=1000),
            AuditReportEntry(filepath='file2.bam', filesize=2000),
            AuditReportEntry(filepath='file3.bam', filesize=None),  # Test None handling
        ]

        with self.assertRaises(TypeError):
            self.reporter.get_report_entries_stats(None)  # type: ignore

        stats = self.reporter.get_report_entries_stats(
            entries[:2]
        )  # Only first two have valid filesize

        self.assertEqual(stats['total_size'], 3000)
        self.assertEqual(stats['file_count'], 2)

    def test_write_csv_report_new_file(self):
        """Test writing CSV report to new file."""
        entries = [
            AuditReportEntry(
                filepath='gs://bucket/file1.bam',
                filesize=1000,
                sg_id='SG01',
                action='DELETE',
            ),
            AuditReportEntry(
                filepath='gs://bucket/file2.bam',
                filesize=2000,
                sg_id='SG02',
                action='REVIEW',
            ),
        ]

        # Mock blob doesn't exist
        mock_blob = unittest.mock.MagicMock()
        mock_blob.exists.return_value = False
        mock_blob.size = 0
        with unittest.mock.patch.object(
            StorageClient, 'get_blob', return_value=mock_blob
        ), unittest.mock.patch.object(StorageClient, 'upload_from_buffer'):
            self.reporter.write_csv_report('test/path.csv', entries, 'TEST REPORT')

            # Verify blob operations
            self.gcs_data_access.storage.get_blob.assert_called_with(  # type: ignore
                'cpg-dataset-main-analysis', 'test/path.csv'
            )
            self.gcs_data_access.storage.upload_from_buffer.assert_called()  # type: ignore

            # Verify the uploaded content format
            call_args = self.gcs_data_access.storage.upload_from_buffer.call_args  # type: ignore
            uploaded_buffer = call_args[0][1]  # Second argument is the buffer
            uploaded_content = uploaded_buffer.getvalue()

            # Should contain headers and data
            self.assertIn('File Path', uploaded_content)
            self.assertIn('gs://bucket/file1.bam', uploaded_content)
            self.assertIn('SG01', uploaded_content)

    def test_write_csv_report_append_to_existing(self):
        """Test appending to existing CSV file."""
        # Mock existing file content
        existing_content = """File Path,File Size,SG ID,Action
gs://bucket/existing.bam,500,SG00,DELETE"""

        mock_blob = unittest.mock.MagicMock()
        mock_blob.exists.return_value = True
        mock_blob.size = 100  # Non-zero size
        mock_blob.download_as_text.return_value = existing_content
        with unittest.mock.patch.object(
            StorageClient, 'get_blob', return_value=mock_blob
        ), unittest.mock.patch.object(StorageClient, 'upload_from_buffer'):
            new_entries = [
                AuditReportEntry(
                    filepath='gs://bucket/new.bam',
                    filesize=1000,
                    sg_id='SG01',
                    action='REVIEW',
                )
            ]
            self.reporter.write_csv_report('test/path.csv', new_entries, 'TEST REPORT')

            # Verify it tried to download existing content
            mock_blob.download_as_text.assert_called_once_with(encoding='utf-8-sig')
            self.gcs_data_access.storage.upload_from_buffer.assert_called()  # type: ignore

    def test_write_audit_reports(self):
        """Test the Reporter.write_audit_reports method."""
        audit_result = AuditResult(
            files_to_delete=[
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file1.bam',
                    filesize=2048000000,
                    sg_id='SG01',
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file4.bam',
                    filesize=1024000000,
                    sg_id='SG02',
                ),
            ],
            files_to_review=[
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file2.bam',
                    filesize=512000000,
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file5.bam',
                    filesize=256000000,
                ),
            ],
            moved_files=[
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file3.bam',
                    filesize=1024000000,
                )
            ],
            unaligned_sequencing_groups=[SEQUENCING_GROUPS['SG05']],
        )
        config = AuditConfig(
            dataset='dataset',
            sequencing_types=('genome', 'exome'),
            sequencing_technologies=('short-read', 'long-read'),
            sequencing_platforms=('illumina', 'pacbio', 'ont'),
            analysis_types=('cram', 'vcf'),
            file_types=tuple(list(FileType)),
            excluded_prefixes=('<EXCLUDED_PREFIX>',),
        )
        with unittest.mock.patch.object(
            Reporter, '_write_csv_report'
        ) as mock_write_csv_report:
            self.reporter.write_audit_reports(audit_result, config)
            stats = self.reporter.generate_summary_statistics(audit_result)
            self.assertIsInstance(stats, dict)
            self.assertEqual(stats['files_to_delete'], 2)
            self.assertEqual(stats['files_to_delete_size_gb'], 3072000000 / (1024**3))
            self.assertEqual(stats['files_to_review'], 2)
            self.assertEqual(stats['files_to_review_size_gb'], 768000000 / (1024**3))
            self.assertEqual(stats['unaligned_sgs'], 1)
            self.assertTrue(mock_write_csv_report.called)

    # ===== REVIEW AUDIT RESULTS TESTS =====
    # Test the report reviewing entrypoint and related functions
    def test_filter_rows(self):
        """Test filtering rows based on expressions."""
        expressions = [
            'sg_type == "exome" and sg_platform == "illumina"',
            'sample_id == "S01" or sample_id == "S02"',
            'participant_external_ids contains "P001"',
            'filesize > 1000 and sg_tech == "short-read"',
            'filepath contains ".bam"',
        ]
        parsed = self.reporter.filter_rows(
            [expressions[0]],
            rows=[
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file1.bam',
                    sg_type='exome',
                    sg_platform='illumina',
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file2.bam',
                    sg_type='genome',
                    sg_platform='illumina',
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file3.bam',
                    sg_type='exome',
                    sg_platform='pacbio',
                ),
            ],
        )
        self.assertEqual(len(parsed), 1)
        self.assertEqual(parsed[0].filepath, 'gs://cpg-dataset-main-upload/file1.bam')
        parsed = self.reporter.filter_rows(
            [expressions[1]],
            rows=[
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file1.bam',
                    sample_id='S01',
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file2.bam',
                    sample_id='S03',
                ),
            ],
        )
        self.assertEqual(len(parsed), 1)
        self.assertEqual(parsed[0].filepath, 'gs://cpg-dataset-main-upload/file1.bam')
        parsed = self.reporter.filter_rows(
            [expressions[2]],
            rows=[
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file1.bam',
                    participant_external_ids=['P001', 'P002'],
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file2.bam',
                    participant_external_ids=['P002', 'P003'],
                ),
            ],
        )
        self.assertEqual(len(parsed), 1)
        self.assertEqual(parsed[0].filepath, 'gs://cpg-dataset-main-upload/file1.bam')
        parsed = self.reporter.filter_rows(
            [expressions[3]],
            rows=[
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file1.bam',
                    filesize=2000,
                    sg_tech='short-read',
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file2.bam',
                    filesize=500,
                    sg_tech='short-read',
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file3.bam',
                    filesize=2000,
                    sg_tech='long-read',
                ),
            ],
        )
        self.assertEqual(len(parsed), 1)
        self.assertEqual(parsed[0].filepath, 'gs://cpg-dataset-main-upload/file1.bam')

        # Now a compound expression
        parsed = self.reporter.filter_rows(
            [expressions[4], 'filesize >= 1500'],
            rows=[
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file1.bam',
                    filesize=2000,
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file2.cram',
                    filesize=1499,
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file3.bam',
                    filesize=500,
                ),
            ],
        )
        self.assertEqual(len(parsed), 1)
        self.assertEqual(parsed[0].filepath, 'gs://cpg-dataset-main-upload/file1.bam')

        # Compound with 'or'
        parsed = self.reporter.filter_rows(
            [
                'filesize >= 1500 or filepath contains ".cram" or sample_id == "S03" or sg_type == "genome"'
            ],
            rows=[
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file1.bam',
                    filesize=2000,
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file2.cram',
                    filesize=1500,
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file3.bam',
                    filesize=500,
                    sg_type='genome',
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file4.bam',
                    sample_id='S03',
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file5.bam',
                    filesize=1000,
                ),
            ],
        )
        self.assertEqual(len(parsed), 4)
        self.assertEqual(parsed[0].filepath, 'gs://cpg-dataset-main-upload/file1.bam')
        self.assertEqual(parsed[1].filepath, 'gs://cpg-dataset-main-upload/file2.cram')
        self.assertEqual(parsed[2].filepath, 'gs://cpg-dataset-main-upload/file3.bam')
        self.assertEqual(parsed[3].filepath, 'gs://cpg-dataset-main-upload/file4.bam')

        parsed = self.reporter.filter_rows(
            ['sample_id != "S03"'],
            rows=[
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file1.bam',
                    sample_id='S01',
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file2.bam',
                    sample_id='S03',
                ),
            ],
        )
        self.assertEqual(len(parsed), 1)
        self.assertEqual(parsed[0].filepath, 'gs://cpg-dataset-main-upload/file1.bam')

    def test_review_rows(self):
        """Test reviewing rows - updating action and comments"""
        rows = [
            AuditReportEntry(
                filepath='gs://cpg-dataset-main-upload/file1.bam',
                review_action='REVIEW',
                review_comment='',
            ),
            AuditReportEntry(
                filepath='gs://cpg-dataset-main-upload/file2.bam',
                review_action='REVIEW',
                review_comment='',
            ),
        ]
        # Mock to_path to return a mock object with exists() method
        mock_path = unittest.mock.MagicMock()
        mock_path.exists.return_value = True
        with unittest.mock.patch(
            'metamist.audit.cli.review_audit_results.to_path', return_value=mock_path
        ):
            review_result = review_rows(
                rows,
                action='INGEST',
                comment='Reviewed and marked for ingestion',
                audit_logs=self.audit_logs,
            )
            self.assertEqual(len(review_result.reviewed_files), 2)
            for row in review_result.reviewed_files:
                self.assertEqual(row.action, 'INGEST')
                self.assertEqual(
                    row.review_comment, 'Reviewed and marked for ingestion'
                )

            rows = [
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file1.bam',
                    review_action='REVIEW',
                    review_comment='',
                ),
                AuditReportEntry(
                    filepath='gs://cpg-dataset-main-upload/file2.bam',
                    review_action='REVIEW',
                    review_comment='',
                ),
            ]
            review_result = review_rows(
                rows,
                action='REVIEW',
                comment='No changes made',
                audit_logs=self.audit_logs,
            )
            self.assertEqual(len(review_result.reviewed_files), 2)
            for row in review_result.reviewed_files:
                self.assertEqual(row.action, 'REVIEW')
                self.assertEqual(row.review_comment, 'No changes made')

            with self.assertRaises(ValueError):
                review_rows(
                    rows,
                    action='INVALID_ACTION',
                    comment='This should fail',
                    audit_logs=self.audit_logs,
                )

    def test_delete_files_from_report(self):
        """Test the delete_files_from_report function end-to-end"""
        rows = [
            AuditReportEntry(
                filepath='gs://cpg-dataset-main-upload/file1.bam',
                filesize=2048000000,
                action='DELETE',
                review_comment='Marked for deletion',
            ),
            AuditReportEntry(
                filepath='gs://cpg-dataset-main-upload/file2.bam',
                filesize=512000000,
                action='REVIEW',
                review_comment='No changes made',
            ),
            AuditReportEntry(
                filepath='gs://cpg-dataset-main-upload/file3.bam',
                filesize=1024000000,
                action='DELETE',
                review_comment='Marked for deletion',
            ),
        ]
        # Mock to_path to return a mock object with exists() method
        mock_path = unittest.mock.MagicMock()
        mock_path.exists.return_value = True
        with unittest.mock.patch(
            'metamist.audit.cli.delete_from_audit_results.to_path',
            return_value=mock_path,
        ):
            # Mock MetamistDataAccess methods
            with unittest.mock.patch.object(
                MetamistDataAccess, 'get_audit_deletion_analysis'
            ) as mock_get_analysis, unittest.mock.patch.object(
                MetamistDataAccess, 'create_audit_deletion_analysis'
            ) as mock_create_analysis, unittest.mock.patch.object(
                MetamistDataAccess, 'update_audit_deletion_analysis'
            ) as mock_update_analysis:
                mock_get_analysis.return_value = None
                mock_create_analysis.return_value = 1
                mock_update_analysis.return_value = 1

                deletion_result = delete_files_from_report(
                    self.gcs_data_access,
                    report_name='files_to_review',
                    rows=rows,
                    audit_logs=self.audit_logs,
                    dry_run=False,
                )

                self.assertEqual(len(deletion_result.deleted_files), 2)

                stats = self.reporter.get_report_entries_stats(
                    deletion_result.deleted_files
                )
                self.assertIsInstance(stats, dict)
                self.assertEqual(stats['total_size'], 2048000000 + 1024000000)
                self.assertEqual(stats['file_count'], 2)



import unittest
from test.testbase import DbIsolatedTest, run_as_sync

from db.python.filters.generic import GenericFilter
from db.python.filters.sample import SampleFilter
from db.python.layers.sample import SampleLayer
from models.models import PRIMARY_EXTERNAL_ORG, SampleUpsertInternal


class TestSample(DbIsolatedTest):
    """Test sample class"""

    # tests run in 'sorted by ascii' order
    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()

        self.slayer = SampleLayer(self.connection)

    @run_as_sync
    async def test_add_sample(self):
        """Test inserting a sample"""
        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                active=True,
                meta={'meta': 'meta ;)'},
            )
        )

        samples = await self.connection.connection.fetch_all(
            'SELECT id, type, meta, project FROM sample'
        )
        self.assertEqual(1, len(samples))
        self.assertEqual(sample.id, samples[0]['id'])

        mapping = await self.slayer.get_sample_id_map_by_external_ids(['Test01'])
        self.assertDictEqual({'Test01': sample.id}, mapping)

    @run_as_sync
    async def test_get_sample(self):
        """Test getting formed sample"""
        meta_dict = {'meta': 'meta ;)'}
        s = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                active=True,
                meta=meta_dict,
            )
        )

        sample = await self.slayer.get_by_id(s.id)

        self.assertEqual('blood', sample.type)
        self.assertDictEqual(meta_dict, sample.meta)

    @run_as_sync
    async def test_query_sample_by_eid(self):
        """Test querying samples by an external ID, and check it's returned"""
        meta_dict = {'meta': 'meta ;)'}
        ex_ids = {PRIMARY_EXTERNAL_ORG: 'Test01', 'external_org': 'ex01'}
        s = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids=ex_ids,
                type='blood',
                active=True,
                meta=meta_dict,
            )
        )

        samples = await self.slayer.query(
            SampleFilter(external_id=GenericFilter(eq='Test01'))
        )
        self.assertEqual(1, len(samples))
        self.assertEqual(s.id, samples[0].id)
        self.assertDictEqual(ex_ids, samples[0].external_ids)

        samples = await self.slayer.query(
            SampleFilter(external_id=GenericFilter(eq='ex01'))
        )
        self.assertEqual(1, len(samples))
        self.assertEqual(s.id, samples[0].id)
        self.assertDictEqual(ex_ids, samples[0].external_ids)

        samples = await self.slayer.query(
            SampleFilter(external_id=GenericFilter(eq='ex02'))
        )
        self.assertEqual(0, len(samples))

    @run_as_sync
    async def test_update_sample(self):
        """Test updating a sample"""
        meta_dict = {'meta': 'meta ;)'}
        s = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                active=True,
                meta=meta_dict,
            )
        )

        new_external_id_dict = {PRIMARY_EXTERNAL_ORG: 'Test02'}
        await self.slayer.upsert_sample(
            SampleUpsertInternal(id=s.id, external_ids=new_external_id_dict)
        )

        sample = await self.slayer.get_by_id(s.id)

        self.assertDictEqual(new_external_id_dict, sample.external_ids)

    @run_as_sync
    async def test_nested_samples_and_query(self):
        """
        Test inserting a sample with nested samples and querying them
        """
        nested_sample = SampleUpsertInternal(
            external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
            type='blood',
            nested_samples=[
                SampleUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'Test02'},
                    type='blood',
                    nested_samples=[
                        SampleUpsertInternal(
                            external_ids={PRIMARY_EXTERNAL_ORG: 'Test03'},
                            type='blood',
                        )
                    ],
                )
            ],
        )

        inserted = (await self.slayer.upsert_samples([nested_sample]))[0]
        first_child = inserted.nested_samples[0]
        children_id = {first_child.id, first_child.nested_samples[0].id}

        # get all
        all_samples = await self.slayer.query(
            SampleFilter(project=GenericFilter(eq=self.project_id))
        )
        self.assertEqual(3, len(all_samples))

        # get only the root
        root_samples = await self.slayer.query(
            SampleFilter(
                project=GenericFilter(eq=self.project_id),
                sample_root_id=GenericFilter(isnull=True),
            )
        )
        self.assertEqual(1, len(root_samples))
        self.assertEqual(root_samples[0].id, inserted.id)
        parentless_samples = await self.slayer.query(
            SampleFilter(
                project=GenericFilter(eq=self.project_id),
                sample_parent_id=GenericFilter(isnull=True),
            )
        )
        self.assertEqual(1, len(parentless_samples))
        self.assertEqual(parentless_samples[0].id, inserted.id)

        # get all children
        children = await self.slayer.query(
            SampleFilter(
                project=GenericFilter(eq=self.project_id),
                sample_root_id=GenericFilter(eq=inserted.id),
            )
        )
        self.assertEqual(2, len(children))
        self.assertSetEqual(children_id, {c.id for c in children})

        # get only first child
        first_child_res = await self.slayer.query(
            SampleFilter(
                project=GenericFilter(eq=self.project_id),
                sample_parent_id=GenericFilter(eq=inserted.id),
            )
        )
        self.assertEqual(1, len(first_child_res))
        self.assertEqual(first_child.id, first_child_res[0].id)

    @run_as_sync
    async def test_deleting_root_sample(self):
        """Test that deleting the root sample cascade deletes the nested samples"""
        nested_sample = SampleUpsertInternal(
            external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
            type='blood',
            nested_samples=[
                SampleUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'Test02'},
                    type='blood',
                    nested_samples=[
                        SampleUpsertInternal(
                            external_ids={PRIMARY_EXTERNAL_ORG: 'Test03'},
                            type='blood',
                        )
                    ],
                )
            ],
        )
        inserted = (await self.slayer.upsert_samples([nested_sample]))[0]
        pre_delete_samples = await self.slayer.query(
            SampleFilter(project=GenericFilter(eq=self.project_id))
        )
        self.assertEqual(3, len(pre_delete_samples))

        # external ids do not have a cascade delete, so we need to delete them first
        await self._connection.execute('DELETE FROM sample_external_id')
        # but nested samples have cascade delete
        await self._connection.execute(
            'DELETE FROM sample WHERE id = :id', {'id': inserted.id}
        )

        post_delete_samples = await self.slayer.query(
            SampleFilter(project=GenericFilter(eq=self.project_id))
        )
        self.assertEqual(0, len(post_delete_samples))


class TestSampleUnwrapping(unittest.TestCase):
    """Test unwrapping nested samples into an ordered list of rows"""

    def test_nested_sample_unwrapping_basic(self):
        """Basic case, one level, a few sub-samples"""
        sample = SampleUpsertInternal(
            id=1,
            nested_samples=[
                SampleUpsertInternal(id=2),
                SampleUpsertInternal(id=3),
                SampleUpsertInternal(id=4),
            ],
        )

        unwrapped = SampleLayer.unwrap_nested_samples([sample])

        self.assertEqual(4, len(unwrapped))

        first_row = unwrapped[0]
        self.assertTupleEqual(
            (None, None, 1),
            (first_row.root, first_row.parent, first_row.sample.id),
        )

        last_row = unwrapped[-1]
        self.assertTupleEqual(
            (1, 1, 4),
            (last_row.root.id, last_row.parent.id, last_row.sample.id),
        )

    def test_nested_sample_unwrapping_many_layers(self):
        """
        Multiple levels, but less than the max depth
        I wrote this explicitly so there couldn't be an issue in a codegen
        """
        sample = SampleUpsertInternal(
            id=1,
            nested_samples=[
                SampleUpsertInternal(
                    id=2,
                    nested_samples=[
                        SampleUpsertInternal(
                            id=3,
                            nested_samples=[
                                SampleUpsertInternal(
                                    id=4,
                                    nested_samples=[
                                        SampleUpsertInternal(
                                            id=5,
                                            nested_samples=[
                                                SampleUpsertInternal(
                                                    id=6,
                                                    nested_samples=[
                                                        SampleUpsertInternal(id=7)
                                                    ],
                                                )
                                            ],
                                        )
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
            ],
        )

        unwrapped = SampleLayer.unwrap_nested_samples([sample])

        self.assertEqual(7, len(unwrapped))

        first_row = unwrapped[0]
        self.assertTupleEqual(
            (None, None, 1),
            (first_row.root, first_row.parent, first_row.sample.id),
        )

        last_row = unwrapped[-1]
        self.assertTupleEqual(
            (1, 6, 7),
            (last_row.root.id, last_row.parent.id, last_row.sample.id),
        )

    def test_nested_sample_unwrapping_overflow(self):
        """
        Test something way too deep, so that the depth protection is triggered
        """
        # at least 20 layers
        root = SampleUpsertInternal(id=1)

        prev = root
        for i in range(2, 21):
            new_sample = SampleUpsertInternal(id=i)
            prev.nested_samples = [new_sample]
            prev = new_sample

        with self.assertRaises(SampleLayer.SampleUnwrapMaxDepthError):
            SampleLayer.unwrap_nested_samples([root])



import socket
from testcontainers.mysql import MySqlContainer


class TestDatabaseContainer:
    """
    Logic to create a singleton MYSQL MariaDB test database container
    """

    _instance = None
    _db_container = None
    _db_port = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(TestDatabaseContainer, cls).__new__(cls)
        return cls._instance

    def _find_free_port(self):
        """Find free port to run tests on"""
        s = socket.socket()
        s.bind(('localhost', 0))  # Bind to a free port provided by the host.
        free_port_number = s.getsockname()[1]  # Return the port number assigned.
        s.close()  # free the port so we can immediately use
        return free_port_number

    def start(self):
        """Set up the test database container"""
        if self._db_container is None:
            self._db_container = MySqlContainer('mariadb:11.2.2', password='test')
            self._db_port = self._find_free_port()
            self._db_container.with_bind_ports(self._db_container.port, self._db_port)
            self._db_container.start()

    def teardown(self):
        """Stop the test database container"""
        if self._db_container is not None:
            self._db_container.stop()
            self._db_container = None
            self._db_port = None

    def get_container(self):
        """Return the test database container"""
        return self._db_container

    def get_port(self):
        """Return the free port number"""
        return self._db_port



# pylint: disable=invalid-overridden-method,no-member

import asyncio
import dataclasses
import logging
import os
import subprocess
import unittest
from functools import wraps
from unittest.mock import patch

import databases.core
import nest_asyncio
from google.auth.credentials import AnonymousCredentials
from google.cloud.storage import Client
from pymysql import IntegrityError
from testcontainers.mysql import MySqlContainer

from api.graphql.loaders import get_context  # type: ignore
from api.graphql.schema import schema  # type: ignore
from db.python.connect import (
    TABLES_ORDERED_BY_FK_DEPS,
    Connection,
    CredentialedDatabaseConfiguration,
    SMConnections,
)
from db.python.tables.project import ProjectPermissionsTable
from models.models.project import Project, ProjectId, ProjectMemberUpdate
from test.testdb_container import TestDatabaseContainer

TEST_PROJECT_NAME = 'test-test'

# use this to determine where the db directory is relatively,
# as pycharm runs in "test/" folder, and GH runs them in git root
am_i_in_test_environment = os.getcwd().endswith('test')

# This monkey patches the asyncio event loop and allows it to be re-entrant
# (you can call run_until_complete while run_until_complete is already on the stack)
# Source: https://stackoverflow.com/a/56434301
nest_asyncio.apply()


for lname in (
    'asyncio',
    'urllib3',
    'docker',
    'databases',
    'testcontainers.core.container',
    'testcontainers.core.waiting_utils',
):
    logging.getLogger(lname).setLevel(logging.WARNING)


loop = asyncio.new_event_loop()


def run_as_sync(f):
    """
    Decorate your async function to run is synchronously
    This runs on the testing event loop
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        return loop.run_until_complete(f(*args, **kwargs))

    return wrapper


class DbTest(unittest.TestCase):
    """Base class for database integration tests"""

    # store connections here, so they can be created PER-CLASS
    # and don't get recreated per test.
    dbs: MySqlContainer | None = None
    connections: dict[str, databases.Database] = {}

    author: str
    project_id: ProjectId
    project_name: str
    project_id_map: dict[ProjectId, Project]
    project_name_map: dict[str, Project]
    db_name: str

    @classmethod
    def setUpClass(cls) -> None:
        @run_as_sync
        async def setup():
            """
            This reuses the mariadb container created before executing any tests.
            Then for each test class it applies liquibase schema and inserts a project.
            """
            # Set environment to test
            os.environ['SM_ENVIRONMENT'] = 'test'
            logger = logging.getLogger()
            try:
                db = cls.dbs
                db_port = None
                if not cls.dbs:
                    logger.disabled = True
                    db_container = TestDatabaseContainer()
                    db = db_container.get_container()
                    db_port = db_container.get_port()
                    logger.disabled = False
                    cls.dbs = db

                db_prefix = 'db'
                if am_i_in_test_environment:
                    db_prefix = '../db'

                if not db:
                    raise ValueError('No database container found')

                # create the database
                db_name = str(cls.__name__) + 'Db'

                _root_connection = SMConnections.make_connection(
                    CredentialedDatabaseConfiguration(
                        host=db.get_container_host_ip(),
                        port=str(db_port),
                        username='root',
                        password=db.password,
                        dbname=db.dbname,
                    ),
                )

                # create the database for each test class, and give permissions
                await _root_connection.connect()
                await _root_connection.execute(f'CREATE DATABASE {db_name};')
                await _root_connection.execute(
                    f"GRANT ALL PRIVILEGES ON `{db_name}`.* TO {db.username}@'%';"
                )
                await _root_connection.execute('FLUSH PRIVILEGES;')
                await _root_connection.disconnect()

                # mfranklin -> future dancoates: if you work out how to copy the
                #       database instead of running liquibase, that would be great
                lcon_string = (
                    f'jdbc:mariadb://{db.get_container_host_ip()}:{db_port}/{db_name}'
                )
                # apply the liquibase schema
                command = [
                    'liquibase',
                    *('--changeLogFile', db_prefix + '/project.xml'),
                    *('--defaultsFile', db_prefix + '/liquibase.properties'),
                    *('--url', lcon_string),
                    *('--driver', 'org.mariadb.jdbc.Driver'),
                    *(
                        '--classpath',
                        db_prefix + '/mariadb-java-client-3.0.3.jar',
                    ),
                    *('--username', db.username),
                    *('--password', db.password),
                    'update',
                ]
                subprocess.check_output(command, stderr=subprocess.STDOUT)

                cls.author = 'testuser'

                sm_db = SMConnections.make_connection(
                    CredentialedDatabaseConfiguration(
                        host=db.get_container_host_ip(),
                        port=str(db_port),
                        username='root',
                        password=db.password,
                        dbname=db_name,
                    )
                )
                await sm_db.connect()

                cls.connections[cls.__name__] = sm_db
                formed_connection = Connection(
                    connection=sm_db,
                    author=cls.author,
                    project_id_map={},
                    project_name_map={},
                    on_behalf_of=None,
                    ar_guid=None,
                    project=None,
                )

                # Add the test user to the admin groups
                await formed_connection.connection.execute(
                    f"""
                    INSERT INTO group_member(group_id, member)
                    SELECT id, '{cls.author}'
                    FROM `group` WHERE name IN('project-creators', 'members-admin')
                """
                )

                ppt = ProjectPermissionsTable(
                    connection=formed_connection,
                )
                cls.project_name = TEST_PROJECT_NAME
                cls.project_id = await ppt.create_project(
                    project_name=cls.project_name,
                    dataset_name='test',
                    author=cls.author,
                )

                _, initial_project_name_map = await ppt.get_projects_accessible_by_user(
                    user=cls.author
                )
                created_project = initial_project_name_map.get(cls.project_name)
                assert created_project

                await ppt.set_project_members(
                    project=created_project,
                    members=[
                        ProjectMemberUpdate(
                            member=cls.author, roles=['reader', 'writer']
                        )
                    ],
                )

                # Get the new project map now that project membership is updated
                (
                    project_id_map,
                    project_name_map,
                ) = await ppt.get_projects_accessible_by_user(user=cls.author)

                cls.project_id_map = project_id_map
                cls.project_name_map = project_name_map
                cls.db_name = db_name

                # formed_connection.project = cls.project_id

            except subprocess.CalledProcessError as e:
                logging.exception(e)
                if e.stderr:
                    logging.error(e.stderr.decode().replace('\\n', '\n'))
                if e.stdout:
                    logging.error(e.stdout.decode().replace('\\n', '\n'))
                cls.tearDownClass()
                raise e
            except Exception as e:
                logging.error(f'FAILED WITH {e}')
                logging.exception(e)
                cls.tearDownClass()

                raise

        return setup()

    @classmethod
    def tearDownClass(cls) -> None:
        # remove from active_tests
        @run_as_sync
        async def tearDown():
            connection = cls.connections.pop(cls.__name__, None)
            if connection:
                await connection.execute(
                    f"""DROP DATABASE IF EXISTS `{cls.db_name}`;"""
                )
                await connection.disconnect()

        tearDown()

    def setUp(self) -> None:
        self._connection = self.connections[self.__class__.__name__]
        # create a connection on each test so we can generate a new
        # audit_log ID for each test
        self.connection = Connection(
            connection=self._connection,
            project=self.project_id_map.get(self.project_id),
            author=self.author,
            project_id_map=self.project_id_map,
            project_name_map=self.project_name_map,
            ar_guid=None,
            on_behalf_of=None,
        )

        # Patch get_gcs_client for all tests
        patcher = patch(
            'models.models.output_file.get_gcs_client',
            self.custom_get_gcs_client,
        )
        self.addCleanup(patcher.stop)
        self.mock_get_gcs_client = patcher.start()

    def custom_get_gcs_client(self):
        """Create the custom client instance with the desired configuration"""
        return Client(
            credentials=AnonymousCredentials(),
            project='test',
            client_options={'api_endpoint': 'http://localhost:4443'},
        )

    async def row_count(self, table: str) -> int:
        """Return the number of rows in the specified table (via raw SQL)"""
        query = f'SELECT COUNT(*) FROM {table}'
        return await self.connection.connection.fetch_val(query)

    @run_as_sync
    async def run_graphql_query(self, query, variables=None):
        """Run SYNC graphql query on internal database"""
        return await self.run_graphql_query_async(query, variables=variables)

    async def run_graphql_query_async(self, query, variables=None) -> dict:
        """Run ASYNC graphql query on internal database"""
        if not isinstance(query, str):
            # if the query was wrapped in a gql() call, unwrap it
            try:
                query = query.loc.source.body
            except KeyError as exc:
                # it obviously wasn't of type document
                raise ValueError(
                    f'Invalid test query (type: {type(query)}): {query}'
                ) from exc

        value = await schema.execute(
            query,
            variable_values=variables,
            context_value=await get_context(
                connection=self.connection,
                request=None,  # pylint: disable
            ),
        )
        if value.errors:
            raise value.errors[0]

        if value.data is None:
            raise ValueError(f'No data returned from query: {query}')

        return value.data

    async def audit_log_id(self):
        """Get audit_log_id for the test"""
        return await self.connection.audit_log_id()

    def assertDataclassEqual(self, a, b):
        """Assert two dataclasses are equal"""

        def to_dict(obj):
            d = dataclasses.asdict(obj)
            for k, v in d.items():
                if dataclasses.is_dataclass(v):
                    d[k] = to_dict(v)
            return d

        self.maxDiff = None
        self.assertDictEqual(to_dict(a), to_dict(b))


class DbIsolatedTest(DbTest):
    """
    Database integration tests that performs clean-up at the start of each test
    """

    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()
        # make sure it's clear, for whatever reason
        await self.clear_database()

    @run_as_sync
    async def tearDown(self) -> None:
        super().tearDown()
        # tear down on end to test any issues with deleting THAT test data
        await self.clear_database()

    async def clear_database(self):
        """Clear the database of all data, except for project + group tables"""
        ignore = {
            'DATABASECHANGELOG',
            'DATABASECHANGELOGLOCK',
            'project',
            'group',
            'project_member',
        }
        for table in TABLES_ORDERED_BY_FK_DEPS:
            if table in ignore:
                continue
            try:
                # mfranklin: Can't use truncate, despite what the docs say
                #   docs: https://mariadb.com/kb/en/truncate-table/
                #   error: System-versioned tables do not support TRUNCATE TABLE'
                #   ticket: https://jira.mariadb.org/browse/MDEV-28439
                #
                # so disable FK checks earlier to more easily delete all rows
                await self.connection.connection.execute(
                    f"""
                    SET GLOBAL FOREIGN_KEY_CHECKS = 0;
                    DELETE FROM `{table}` WHERE 1;
                    SET GLOBAL FOREIGN_KEY_CHECKS = 1;
                    """
                )
                await self.connection.connection.execute(
                    f'DELETE HISTORY FROM `{table}`'
                )
            except IntegrityError as e:
                raise IntegrityError(f'Could not delete {table}') from e



"""
Metamist Audit Module

This module provides tools for auditing Metamist datasets and their associated
cloud storage buckets. The primary use case is identifying:
- Files that can be safely deleted (e.g., original reads where CRAMs exist)
- Files that need review (could require ingestion into Metamist, could require deletion)
- Sequencing groups requiring processing

Architecture:
- models/: Data models and value objects
- adapters/: External interface adapters (GraphQL, GCS)
- data_access/: Data access layer
- services/: Business logic
- cli/: Command-line interfaces

Usage:
    from metamist.audit import audit_upload_bucket, AuditConfig

    config = AuditConfig(
        dataset='my-dataset',
        sequencing_types=('genome',),
        sequencing_technologies=('short-read',),
        sequencing_platforms=('illumina',),
        analysis_types=('CRAM',),
        file_types=(FileType.FASTQ, FileType.BAM),
    )

    audit_upload_bucket(config)
"""

from .models import (
    # Core entities
    SequencingGroup,
    Analysis,
    Assay,
    Sample,
    Participant,
    AuditReportEntry,
    AuditResult,
    ReviewResult,
    DeletionResult,
    # Value objects
    AuditConfig,
    FileType,
    FileMetadata,
    ExternalIds,
    MovedFile,
)

from .services import (
    BucketAuditLogger,
    AuditAnalyzer,
    Reporter,
    FileMatchingService,
)

from .cli.upload_bucket_audit import (
    audit_upload_bucket,
    audit_upload_bucket_async,
    AuditOrchestrator,
)
from .cli.review_audit_results import review_audit_report, review_rows
from .cli.delete_from_audit_results import (
    delete_from_audit_results,
    delete_files_from_report,
    upsert_deleted_files_analysis,
)

__version__ = '2.0.0'

__all__ = [
    # Core functions
    'audit_upload_bucket',
    'audit_upload_bucket_async',
    'AuditOrchestrator',
    'review_audit_report',
    'review_rows',
    'delete_from_audit_results',
    'delete_files_from_report',
    'upsert_deleted_files_analysis',
    # Configuration
    'AuditConfig',
    # Enums
    'FileType',
    # Core entities
    'SequencingGroup',
    'Analysis',
    'Assay',
    'Sample',
    'Participant',
    # Results
    'AuditResult',
    'AuditReportEntry',
    'ReviewResult',
    'DeletionResult',
    # Services
    'AuditAnalyzer',
    'BucketAuditLogger',
    'Reporter',
    'FileMatchingService',
    # Value objects
    'FileMetadata',
    'ExternalIds',
    'MovedFile',
]



# pylint: disable=global-statement,import-self
"""
GraphQL utilities for Metamist, allows you to:
    - construct queries using the `gql` function (which validates graphql syntax)
    - validate queries with metamist schema (by fetching the schema)
"""

import os
from json.decoder import JSONDecodeError
from typing import Any, Dict

import backoff
from gql import Client
from gql import gql as gql_constructor
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.aiohttp import log as aiohttp_logger
from gql.transport.exceptions import TransportServerError
from gql.transport.requests import RequestsHTTPTransport
from gql.transport.requests import log as requests_logger

# this does not import itself, it imports the module
from graphql import DocumentNode  # type: ignore
from requests.exceptions import HTTPError

from cpg_utils.cloud import get_google_identity_token

import metamist.configuration

_sync_client: Client | None = None
_async_client: Client | None = None


def get_local_schema() -> str:
    """Get the local schema"""
    with open(os.path.join(os.path.dirname(__file__), 'schema.graphql')) as f:
        contents = f.read()

    return contents


def get_sm_url():
    """Get the URL for the Metamist GraphQL API"""
    return os.path.join(metamist.configuration.sm_url, 'graphql')


def configure_sync_client(
    url: str = None,
    schema: str = None,
    auth_token: str = None,
    force_recreate=False,
    use_local_schema: bool = False,
):
    """Get sync gql Client"""
    global _sync_client

    if use_local_schema:
        # for local schema, we basically don't allow requests
        schema = get_local_schema()
        return Client(schema=schema)

    if _sync_client and not force_recreate:
        return _sync_client

    env = os.getenv('SM_ENVIRONMENT', 'PRODUCTION').lower()
    if env == 'local':
        transport = RequestsHTTPTransport(url=url or get_sm_url())
    else:
        token = auth_token or get_google_identity_token(
            target_audience=metamist.configuration.sm_url
        )
        transport = RequestsHTTPTransport(
            url=url or get_sm_url(),
            headers={'Authorization': f'Bearer {token}'},
        )

    _sync_client = Client(
        transport=transport, schema=schema, fetch_schema_from_transport=schema is None
    )
    # _sync_client.connect_sync()

    return _sync_client


async def configure_async_client(
    url: str = None, schema: str = None, auth_token: str = None, force_recreate=False
) -> Client:
    """Configure an async client for use with the Metamist GraphQL API"""
    global _async_client

    if _async_client and not force_recreate:
        return _async_client

    env = os.getenv('SM_ENVIRONMENT', 'PRODUCTION').lower()
    if env == 'local':
        transport = AIOHTTPTransport(url=url or get_sm_url())
    else:
        token = auth_token or get_google_identity_token(
            target_audience=metamist.configuration.sm_url
        )
        transport = AIOHTTPTransport(
            url=url or get_sm_url(),
            headers={'Authorization': f'Bearer {token}'},
        )

    _async_client = Client(
        transport=transport, schema=schema, fetch_schema_from_transport=schema is None
    )
    # await _async_client.connect_async()
    return _async_client


async def get_async_client():
    """Get async gql Client"""
    return await configure_async_client()


def gql(
    request_string: str, should_validate=False, use_local_schema=False
) -> DocumentNode:
    """
    Given a String containing a GraphQL request, parse it into a Document.
    Optionally validate by fetching schema from metamist at SM_URL / SM_ENVIRONMENT
    """
    doc = gql_constructor(request_string)
    if should_validate:
        validate(doc, use_local_schema=use_local_schema)
    return doc


def validate(doc: DocumentNode, client=None, use_local_schema=False):
    """
    Validate a GraphQL document against the schema.
    Fetch schema from SM_URL / SM_ENVIRONMENT
    """
    if client is not None and use_local_schema:
        raise ValueError('Cannot provide local_schema and client to validate')
    if isinstance(doc, str):
        doc = gql(doc)

    (client or configure_sync_client(use_local_schema=use_local_schema)).validate(doc)
    return True


# use older style typing to broaden supported Python versions
@backoff.on_exception(
    backoff.expo,
    factor=3,
    jitter=None,
    exception=(HTTPError, JSONDecodeError, TransportServerError),
    max_tries=5,
    max_time=60,
)
def query(
    _query: str | DocumentNode,
    variables: Dict | None = None,
    client: Client | None = None,
    log_response: bool = False,
) -> Dict[str, Any]:
    """Query the metamist GraphQL API"""
    if variables is None:
        variables = {}

    # disable logging for gql
    current_level = aiohttp_logger.level
    if not log_response:
        requests_logger.setLevel('WARNING')

    response = (client or configure_sync_client()).execute_sync(
        _query if isinstance(_query, DocumentNode) else gql(_query),
        variable_values=variables,
    )

    if not log_response:
        requests_logger.setLevel(current_level)
    return response


@backoff.on_exception(
    backoff.expo,
    factor=3,
    jitter=None,
    exception=(HTTPError, JSONDecodeError, TransportServerError),
    max_tries=5,
    max_time=60,
)
async def query_async(
    _query: str | DocumentNode,
    variables: Dict | None = None,
    client: Client | None = None,
    log_response: bool = False,
) -> Dict[str, Any]:
    """Asynchronously query the Metamist GraphQL API"""
    if variables is None:
        variables = {}

    # disable logging for gql
    current_level = aiohttp_logger.level
    if log_response:
        aiohttp_logger.setLevel('WARNING')

    if not client:
        client = await configure_async_client()

    response = await client.execute_async(
        _query if isinstance(_query, DocumentNode) else gql(_query),
        variable_values=variables,
    )

    if log_response:
        aiohttp_logger.setLevel(current_level)

    return response



from collections import defaultdict
from test.testbase import DbIsolatedTest, run_as_sync

from pymysql.err import IntegrityError

from db.python.enum_tables import AssayTypeTable
from db.python.filters import GenericFilter
from db.python.layers.assay import AssayLayer
from db.python.layers.sample import SampleLayer
from db.python.tables.assay import AssayFilter, AssayTable
from db.python.utils import NotFoundError
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    AssayUpsertInternal,
    SampleUpsertInternal,
)
from models.models.sequencing_group import SequencingGroupUpsertInternal

default_sequencing_meta = {
    'sequencing_type': 'genome',
    'sequencing_platform': 'short-read',
    'sequencing_technology': 'illumina',
}


class TestAssay(DbIsolatedTest):
    """Test assay class"""

    # pylint: disable=too-many-instance-attributes

    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()

        self.slayer = SampleLayer(self.connection)
        self.assaylayer = AssayLayer(self.connection)
        self.external_sample_id = 'TESTING001'

        # Create new sample
        self.sample_id_raw = (
            await self.slayer.upsert_sample(
                SampleUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: self.external_sample_id},
                    type='blood',
                    active=True,
                    meta={'Testing': 'test_assay'},
                )
            )
        ).id

        at = AssayTypeTable(self.connection)
        await at.insert('metabolomics')

    @run_as_sync
    async def test_not_found_assay(self):
        """
        Test the NotFoundError when getting an invalid assay ID
        """

        @run_as_sync
        async def get():
            return await self.assaylayer.get_assay_by_id(-1)

        self.assertRaises(NotFoundError, get)

    @run_as_sync
    async def test_upsert_assay(self):
        """
        Test inserting a assay, and check all values are inserted correctly
        """
        external_ids = {'default': 'SEQ01', 'collaborator2': 'CBSEQ_1'}
        meta = {
            '1': 1,
            'nested': {'nested': 'dict'},
            'alpha': ['b', 'e', 't'],
            **default_sequencing_meta,
        }
        upserted_assay = await self.assaylayer.upsert_assay(
            AssayUpsertInternal(
                sample_id=self.sample_id_raw,
                type='sequencing',
                meta=meta,
                external_ids=external_ids,
            )
        )

        assay = await self.assaylayer.get_assay_by_id(assay_id=upserted_assay.id)

        self.assertEqual(upserted_assay.id, assay.id)
        self.assertEqual(self.sample_id_raw, int(assay.sample_id))
        self.assertEqual('sequencing', assay.type)
        self.assertDictEqual(external_ids, assay.external_ids)
        self.assertDictEqual(meta, assay.meta)

    @run_as_sync
    async def test_insert_assays_for_each_type(self):
        """
        Test inserting a assay, and check all values are inserted correctly
        """
        meta = {
            '1': 1,
            'nested': {'nested': 'dict'},
            'alpha': ['b', 'e', 't'],
            **default_sequencing_meta,
        }
        sequencing_types = ['sequencing', 'metabolomics']
        assays = await self.assaylayer.upsert_assays(
            [
                AssayUpsertInternal(
                    external_ids={'eid': f'external_id_{_type}'},
                    sample_id=self.sample_id_raw,
                    type='sequencing',
                    meta=meta,
                )
                for _type in sequencing_types
            ]
        )
        assay_ids = [a.id for a in assays]
        inserted_types_rows = await self.connection.connection.fetch_all(
            'SELECT type FROM assay WHERE id in :ids', {'ids': assay_ids}
        )
        inserted_types = set(r['type'] for r in inserted_types_rows)

        self.assertEqual(len(sequencing_types), len(assay_ids))
        self.assertEqual(1, len(inserted_types))

    @run_as_sync
    async def test_clashing_external_ids(self):
        """Test that should fail when 2nd assay is inserted with same external_id"""
        external_ids = {'default': 'clashing'}
        await self.assaylayer.upsert_assay(
            AssayUpsertInternal(
                sample_id=self.sample_id_raw,
                type='sequencing',
                meta={**default_sequencing_meta},
                external_ids=external_ids,
            )
        )

        @run_as_sync
        async def _insert_failing_assay():
            return await self.assaylayer.upsert_assay(
                AssayUpsertInternal(
                    sample_id=self.sample_id_raw,
                    type='sequencing',
                    meta={**default_sequencing_meta},
                    external_ids=external_ids,
                )
            )

        _n_assays_query = 'SELECT COUNT(*) from assay'
        self.assertEqual(1, await self.connection.connection.fetch_val(_n_assays_query))
        self.assertRaises(IntegrityError, _insert_failing_assay)
        # make sure the transaction unwinds the insert second assay if the external_id clashes
        self.assertEqual(1, await self.connection.connection.fetch_val(_n_assays_query))

    @run_as_sync
    async def test_insert_clashing_external_ids_multiple(self):
        """
        Test inserting a assay, and check all values are inserted correctly
        """
        external_ids = {'default': 'clashing'}

        @run_as_sync
        async def _insert_clashing():
            return await self.assaylayer.upsert_assays(
                [
                    AssayUpsertInternal(
                        # both get the same external_ids
                        external_ids=external_ids,
                        sample_id=self.sample_id_raw,
                        type='sequencing',
                        meta={**default_sequencing_meta},
                    )
                    for _ in range(2)
                ]
            )

        _n_assays_query = 'SELECT COUNT(*) from assay'

        self.assertEqual(0, await self.connection.connection.fetch_val(_n_assays_query))
        self.assertRaises(IntegrityError, _insert_clashing)
        self.assertEqual(0, await self.connection.connection.fetch_val(_n_assays_query))

    @run_as_sync
    async def test_getting_assay_by_external_id(self):
        """
        Test get differences assays by multiple IDs
        """
        seq1 = await self.assaylayer.upsert_assay(
            AssayUpsertInternal(
                sample_id=self.sample_id_raw,
                type='sequencing',
                meta={**default_sequencing_meta},
                external_ids={'default': 'SEQ01', 'other': 'EXT_SEQ1'},
            )
        )
        seq2 = await self.assaylayer.upsert_assay(
            AssayUpsertInternal(
                sample_id=self.sample_id_raw,
                type='sequencing',
                meta={**default_sequencing_meta},
                external_ids={'default': 'SEQ02'},
            )
        )

        fquery_1 = AssayFilter(
            external_id=GenericFilter(eq='SEQ01'),
            project=GenericFilter(eq=self.project_id),
        )
        self.assertEqual(seq1.id, (await self.assaylayer.query(fquery_1))[0].id)
        fquery_2 = AssayFilter(
            external_id=GenericFilter(eq='EXT_SEQ1'),
            project=GenericFilter(eq=self.project_id),
        )
        self.assertEqual(seq1.id, (await self.assaylayer.query(fquery_2))[0].id)
        fquery_3 = AssayFilter(
            external_id=GenericFilter(eq='SEQ02'),
            project=GenericFilter(eq=self.project_id),
        )
        self.assertEqual(seq2.id, (await self.assaylayer.query(fquery_3))[0].id)

    @run_as_sync
    async def test_query(self):
        """Test query_assays in different combinations"""
        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'SAM_TEST_QUERY'},
                type='blood',
                active=True,
                meta={'collection-year': '2022'},
            )
        )

        sample_id_for_test = sample.id

        assert sample_id_for_test is not None, 'Sample ID is None'

        seqs = await self.assaylayer.upsert_assays(
            [
                AssayUpsertInternal(
                    sample_id=sample_id_for_test,
                    type='sequencing',
                    meta={'unique': 'a', 'common': 'common', **default_sequencing_meta},
                    external_ids={'default': 'SEQ01'},
                ),
                AssayUpsertInternal(
                    sample_id=sample_id_for_test,
                    type='sequencing',
                    meta={'unique': 'b', 'common': 'common', **default_sequencing_meta},
                    external_ids={'default': 'SEQ02'},
                ),
            ]
        )

        async def search_result_to_ids(filter_: AssayFilter):
            filter_.project = GenericFilter(eq=self.project_id)
            seqs = await self.assaylayer.query(filter_)
            return {s.id for s in seqs}

        seq1_id = seqs[0].id
        seq2_id = seqs[1].id

        # sample_ids
        self.assertSetEqual(
            {seq1_id, seq2_id},
            await search_result_to_ids(
                AssayFilter(sample_id=GenericFilter(in_=[sample_id_for_test]))
            ),
        )
        self.assertSetEqual(
            set(),
            await search_result_to_ids(
                AssayFilter(sample_id=GenericFilter(in_=[9_999_999]))
            ),
        )

        # external assay IDs
        self.assertSetEqual(
            {seq1_id},
            await search_result_to_ids(
                AssayFilter(external_id=GenericFilter(eq='SEQ01'))
            ),
        )
        self.assertSetEqual(
            {seq1_id, seq2_id},
            await search_result_to_ids(
                AssayFilter(
                    external_id=GenericFilter(in_=['SEQ01', 'SEQ02']),
                )
            ),
        )

        # seq_meta
        self.assertSetEqual(
            {seq2_id},
            await search_result_to_ids(
                AssayFilter(meta={'unique': GenericFilter(eq='b')})
            ),
        )
        self.assertSetEqual(
            {seq1_id, seq2_id},
            await search_result_to_ids(
                AssayFilter(meta={'common': GenericFilter(eq='common')})
            ),
        )

        # sample meta
        self.assertSetEqual(
            {seq1_id, seq2_id},
            await search_result_to_ids(
                AssayFilter(sample_meta={'collection-year': GenericFilter(eq='2022')})
            ),
        )
        self.assertSetEqual(
            set(),
            await search_result_to_ids(
                AssayFilter(sample_meta={'unknown_key': GenericFilter(eq='2022')})
            ),
        )

        # assay types
        self.assertSetEqual(
            {seq1_id, seq2_id},
            await search_result_to_ids(
                AssayFilter(type=GenericFilter(in_=['sequencing']))
            ),
        )

        # combination
        self.assertSetEqual(
            {seq2_id},
            await search_result_to_ids(
                AssayFilter(
                    sample_meta={'collection-year': GenericFilter(eq='2022')},
                    external_id=GenericFilter(in_=['SEQ02']),
                )
            ),
        )
        self.assertSetEqual(
            {seq1_id},
            await search_result_to_ids(
                AssayFilter(
                    external_id=GenericFilter(in_=['SEQ01']),
                    type=GenericFilter(eq='sequencing'),
                )
            ),
        )
        # self.assertSetEqual(
        #     set(),
        #     await search_result_to_ids(
        #         external_assay_ids=['SEQ01'], types=[SequenceType.EXOME]
        #     ),
        # )

    @run_as_sync
    async def test_query_by_sg_ids(self):
        """Test query_assays in different combinations"""
        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'SAM_TEST_QUERY'},
                type='blood',
                active=True,
                meta={'collection-year': '2022'},
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        meta={'sgmeta': 'sgvalue'},
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                external_ids={'default': 'A1_1'},
                                meta={
                                    'batch': 'batch-1a',
                                    'sequencing_type': 'genome',
                                    'sequencing_platform': 'illumina',
                                    'sequencing_technology': 'short-read',
                                },
                            ),
                            AssayUpsertInternal(
                                type='sequencing',
                                external_ids={'default': 'A1_2'},
                                meta={
                                    'batch': 'batch-1b',
                                    'sequencing_type': 'genome',
                                    'sequencing_platform': 'illumina',
                                    'sequencing_technology': 'short-read',
                                },
                            ),
                        ],
                    ),
                    SequencingGroupUpsertInternal(
                        type='exome',
                        technology='short-read',
                        platform='illumina',
                        meta={'sgmeta': 'sgvalue'},
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                external_ids={'default': 'sg2_1'},
                                meta={
                                    'batch': 'batch-2',
                                    'sequencing_type': 'genome',
                                    'sequencing_platform': 'illumina',
                                    'sequencing_technology': 'short-read',
                                },
                            ),
                            AssayUpsertInternal(
                                type='sequencing',
                                external_ids={'default': 'sg2_2'},
                                meta={
                                    'batch': 'batch-2',
                                    'sequencing_type': 'genome',
                                    'sequencing_platform': 'illumina',
                                    'sequencing_technology': 'short-read',
                                },
                            ),
                        ],
                    ),
                ],
            )
        )

        sg_id = sample.sequencing_groups[0].id
        assay_ids_sg1 = {a.id for a in sample.sequencing_groups[0].assays}

        assays = await self.assaylayer.get_assays_for_sequencing_group_ids([sg_id])

        self.assertSetEqual(
            assay_ids_sg1, {a.id for sgs_as in assays.values() for a in sgs_as}
        )

        # subfilter
        assays_batch_1a = await self.assaylayer.get_assays_for_sequencing_group_ids(
            [sg_id],
            filter_=AssayFilter(
                meta={'batch': GenericFilter(eq='batch-1a')},
            ),
        )
        self.assertEqual(1, len(assays_batch_1a))
        batch_1a_assay = next(iter(assays_batch_1a.values()))[0]
        self.assertEqual('batch-1a', batch_1a_assay.meta['batch'])

    @run_as_sync
    async def test_update(self):
        """Test updating an assay, and all fields are updated correctly"""
        # insert
        assay = await self.assaylayer.upsert_assay(
            AssayUpsertInternal(
                sample_id=self.sample_id_raw,
                type='sequencing',
                meta={'a': 1, 'b': 2, **default_sequencing_meta},
                external_ids={
                    'default': 'SEQ01',
                    'untouched': 'UTC+1',
                    'to_delete': 'VALUE',
                },
            )
        )

        await self.assaylayer.upsert_assay(
            AssayUpsertInternal(
                id=assay.id,
                sample_id=self.sample_id_raw,
                external_ids={
                    'default': 'NSQ_01',
                    'ext': 'EXTSEQ01',
                    'to_delete': None,
                },
                meta={'a': 2, 'c': True},
            )
        )

        update_assay = await self.assaylayer.get_assay_by_id(assay_id=assay.id)

        self.assertEqual(assay.id, update_assay.id)
        self.assertEqual(self.sample_id_raw, int(update_assay.sample_id))
        self.assertEqual('sequencing', update_assay.type)
        self.assertDictEqual(
            {'default': 'NSQ_01', 'ext': 'EXTSEQ01', 'untouched': 'UTC+1'},
            update_assay.external_ids,
        )
        self.assertDictEqual(
            {'a': 2, 'b': 2, 'c': True, **default_sequencing_meta}, update_assay.meta
        )

    @run_as_sync
    async def test_update_type(self):
        """
        Test update all assay statuses
        """
        assay = await self.assaylayer.upsert_assay(
            AssayUpsertInternal(
                sample_id=self.sample_id_raw,
                type='sequencing',
                meta={**default_sequencing_meta},
                external_ids={},
            )
        )

        # cycle through all statuses, and check that works
        await self.assaylayer.upsert_assay(
            AssayUpsertInternal(id=assay.id, type='metabolomics')
        )
        row_to_check = await self.connection.connection.fetch_one(
            'SELECT type FROM assay WHERE id = :id',
            {'id': assay.id},
        )
        self.assertEqual('metabolomics', row_to_check['type'])

    @run_as_sync
    async def test_batch_statistics(self):
        """
        Test batch statistics
        """
        samples_to_insert = [
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'SAMPLE_1'},
                type='blood',
                active=True,
                meta={'collection-year': '2022'},
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        external_ids={'default': 'SG_1'},
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        meta={'sgmeta': 'sgvalue'},
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                external_ids={'default': 'A1_1'},
                                meta={
                                    'batch': 'batch-1',
                                    'sequencing_type': 'genome',
                                    'sequencing_platform': 'illumina',
                                    'sequencing_technology': 'short-read',
                                },
                            ),
                            AssayUpsertInternal(
                                type='sequencing',
                                external_ids={'default': 'A1_2'},
                                meta={
                                    'batch': 'batch-1',
                                    'sequencing_type': 'genome',
                                    'sequencing_platform': 'illumina',
                                    'sequencing_technology': 'short-read',
                                },
                            ),
                        ],
                    ),
                    SequencingGroupUpsertInternal(
                        external_ids={'default': 'SG_2'},
                        type='exome',
                        technology='short-read',
                        platform='illumina',
                        meta={'sgmeta': 'sgvalue'},
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                external_ids={'default': 'A2_1'},
                                meta={
                                    'batch': 'batch-1',
                                    'sequencing_type': 'exome',
                                    'sequencing_platform': 'illumina',
                                    'sequencing_technology': 'short-read',
                                },
                            ),
                            AssayUpsertInternal(
                                type='sequencing',
                                external_ids={'default': 'A2_2'},
                                meta={
                                    'batch': 'batch-2',
                                    'sequencing_type': 'exome',
                                    'sequencing_platform': 'illumina',
                                    'sequencing_technology': 'short-read',
                                },
                            ),
                        ],
                    ),
                ],
            ),
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'SAMPLE_2'},
                type='blood',
                active=True,
                meta={'collection-year': '2022'},
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        external_ids={'default': 'SG_3'},
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        meta={'sgmeta': 'sgvalue'},
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                external_ids={'default': 'A3_1'},
                                meta={
                                    'batch': 'batch-1',
                                    'sequencing_type': 'genome',
                                    'sequencing_platform': 'illumina',
                                    'sequencing_technology': 'short-read',
                                },
                            ),
                            AssayUpsertInternal(
                                type='sequencing',
                                external_ids={'default': 'A3_2'},
                                meta={
                                    'batch': 'batch-1',
                                    'sequencing_type': 'genome',
                                    'sequencing_platform': 'illumina',
                                    'sequencing_technology': 'short-read',
                                },
                            ),
                        ],
                    ),
                    SequencingGroupUpsertInternal(
                        external_ids={'default': 'SG_4'},
                        type='transcriptome',
                        technology='short-read',
                        platform='illumina',
                        meta={'sgmeta': 'sgvalue'},
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                external_ids={'default': 'A4_1'},
                                meta={
                                    'batch': 'batch-3',
                                    'sequencing_type': 'transcriptome',
                                    'sequencing_platform': 'illumina',
                                    'sequencing_technology': 'short-read',
                                },
                            ),
                            AssayUpsertInternal(
                                type='sequencing',
                                external_ids={'default': 'A4_2'},
                                meta={
                                    'batch': 'batch-3',
                                    'sequencing_type': 'transcriptome',
                                    'sequencing_platform': 'illumina',
                                    'sequencing_technology': 'short-read',
                                },
                            ),
                        ],
                    ),
                ],
            ),
        ]

        await SampleLayer(self.connection).upsert_samples(samples_to_insert)
        assay_layer = AssayTable(self.connection)
        rows = await assay_layer.get_assay_type_numbers_by_batch_for_project(
            self.project_id
        )

        assays_in_batch: dict[str, int] = defaultdict(int)
        sgs_in_seq_type_batch: dict[str, dict[str, int]] = defaultdict(
            lambda: defaultdict(int)
        )
        for r in rows:
            # sequencing_group_ids has format [(sg_id, assay_count), ...]
            assays_in_batch[r.batch] += sum(sg[1] for sg in r.sequencing_group_ids)
            sgs_in_seq_type_batch[r.batch][r.sequencing_type] += len(
                r.sequencing_group_ids
            )

        # aggregate to test, mostly to show there are multiple ways you
        # may want to aggregate this information
        self.assertDictEqual(
            assays_in_batch,
            {
                'batch-1': 5,
                'batch-2': 1,
                'batch-3': 2,
            },
        )

        self.assertDictEqual(
            sgs_in_seq_type_batch,
            {
                'batch-1': {'genome': 2, 'exome': 1},
                'batch-2': {'exome': 1},
                'batch-3': {'transcriptome': 1},
            },
        )



"""Report generation service for audit results."""

import csv
from datetime import datetime
from io import StringIO

from cpg_utils import Path, to_path

from .audit_logging import BucketAuditLogger
from metamist.audit.data_access import GCSDataAccess
from metamist.audit.models import (
    AuditResult,
    ReviewResult,
    DeletionResult,
    AuditReportEntry,
    SequencingGroup,
    AuditConfig,
)


class Reporter:
    """Service for generating and writing audit reports."""

    def __init__(
        self,
        gcs_data: GCSDataAccess,
        audit_logs: BucketAuditLogger,
        timestamp: str | None = None,
    ):
        """
        Initialize the report generator.

        Args:
            storage_client: Google Cloud Storage client
            timestamp: Optional timestamp for report file prefix
            audit_logs: Audit logs instance
        """
        self.gcs_data = gcs_data
        self.audit_logs = audit_logs
        self.timestamp = timestamp or datetime.now().strftime('%Y-%m-%d_%H%M%S')
        self.output_dir = f'audit_results/{self.timestamp}'

    # Reading reports #
    def get_report_rows(self, report_path: Path) -> list[AuditReportEntry]:
        """Retrieve rows from the audit report CSV."""
        if not report_path.exists():
            self.audit_logs.error(f'Report file {report_path} does not exist.')
            return []

        with report_path.open('r') as f:
            reader = csv.DictReader(f, fieldnames=AuditReportEntry().fieldnames())
            next(reader)  # Skip header row
            rows = [AuditReportEntry(**row) for row in reader]

        return rows

    def get_report_rows_from_name(self, name: str) -> list[AuditReportEntry]:
        """Retrieve rows from the audit report CSV by name."""
        report_path = to_path(
            f'gs://{self.gcs_data.analysis_bucket}/audit_results/{self.timestamp}/{name}.csv'
        )
        return self.get_report_rows(report_path)

    def get_report_stats(self, report_path: Path) -> dict:
        """Count the rows and sum the total file size for all rows in the report"""
        rows = self.get_report_rows(report_path)
        return self.get_report_entries_stats(rows)

    def get_report_entries_stats(self, entries: list[AuditReportEntry]) -> dict:
        """Get statistics for the given report entries."""
        total_size = sum(int(entry.filesize) or 0 for entry in entries)
        return {'total_size': total_size, 'file_count': len(entries)}

    # Writing reports #
    def write_audit_reports(
        self,
        audit_result: AuditResult,
        config: AuditConfig,
    ):
        """
        Write all audit reports to GCS.

        Args:
            audit_result: The audit analysis results
            bucket_name: GCS bucket name for reports
            config: Audit configuration
            output_prefix: Optional prefix for output paths
        """
        self.audit_logs.info_nl(
            f'Target: gs://{self.gcs_data.analysis_bucket}/{self.output_dir}/'
        )

        self._write_audit_config(f'{self.output_dir}/audit_config.txt', config)

        if audit_result.files_to_delete:
            self._write_csv_report(
                f'{self.output_dir}/files_to_delete.csv',
                audit_result.files_to_delete,
                'FILES TO DELETE',
            )

        if audit_result.files_to_review:
            self._write_csv_report(
                f'{self.output_dir}/files_to_review.csv',
                audit_result.files_to_review,
                'FILES TO REVIEW',
            )

        if audit_result.moved_files:
            self._write_csv_report(
                f'{self.output_dir}/moved_files.csv',
                audit_result.moved_files,
                'MOVED FILES',
            )

        if audit_result.unaligned_sequencing_groups:
            self._write_unaligned_sgs_report(
                f'{self.output_dir}/unaligned_sgs.tsv',
                audit_result.unaligned_sequencing_groups,
            )

    def _write_audit_config(self, blob_path: str, config: AuditConfig):
        """Write metadata about the audit config used in this run to a text file."""
        buffer = StringIO()

        lines = [
            f'Dataset: {config.dataset}',
            f"Sequencing Types: {', '.join(sorted(config.sequencing_types))}",
            f"Sequencing Technologies: {', '.join(sorted(config.sequencing_technologies))}",
            f"Sequencing Platforms: {', '.join(sorted(config.sequencing_platforms))}",
            f"File Types: {', '.join(sorted(ft.name for ft in config.file_types))}",
            f"Analysis Types: {', '.join(sorted(config.analysis_types))}",
        ]
        if config.excluded_prefixes:
            lines.append(
                f"Excluded Prefixes: {', '.join(sorted(config.excluded_prefixes))}"
            )
        lines.append('')

        buffer.write('\n'.join(lines))

        blob = self.gcs_data.storage.get_blob(self.gcs_data.analysis_bucket, blob_path)
        self.gcs_data.storage.upload_from_buffer(blob, buffer)

        self.audit_logs.info_nl(
            f'{"AUDIT CONFIG":<20}: gs://{self.gcs_data.analysis_bucket}/{blob_path}'
        )

    def _write_csv_report(
        self,
        blob_path: str,
        entries: list[AuditReportEntry],
        report_name: str,
    ) -> Path:
        """Create or update a CSV report for audit entries."""
        rows = [entry.to_report_dict() for entry in entries]
        if report_name == 'FILES TO DELETE':
            rows = sorted(rows, key=lambda x: x.get('SG ID', ''))

        blob = self.gcs_data.storage.get_blob(self.gcs_data.analysis_bucket, blob_path)
        output_path = f'gs://{self.gcs_data.analysis_bucket}/{blob_path}'
        if blob.exists() and blob.size > 0:
            self.audit_logs.info_nl(
                f'Existing report found, appending new rows to: {output_path}'
            )
            existing_data = blob.download_as_text(encoding='utf-8-sig')
            buffer = StringIO(existing_data)
            reader = csv.DictReader(buffer)
            existing_rows = list(reader)
            # Prevent duplicates
            rows = existing_rows + [row for row in rows if row not in existing_rows]
            buffer.close()

        buffer = StringIO()
        writer = csv.DictWriter(buffer, fieldnames=AuditReportEntry().fieldnames())
        writer.writeheader()
        writer.writerows(rows)

        blob = self.gcs_data.storage.get_blob(self.gcs_data.analysis_bucket, blob_path)
        self.gcs_data.storage.upload_from_buffer(blob, buffer, 'text/csv')

        self.audit_logs.info_nl(
            f'{report_name:<20}: {output_path} ({len(entries)} entries)'
        )

        return to_path(output_path)

    def write_csv_report(
        self,
        blob_path: str,
        entries: list[AuditReportEntry],
        report_name: str,
    ) -> Path:
        """Write a CSV report to GCS."""
        return self._write_csv_report(
            blob_path=blob_path,
            entries=entries,
            report_name=report_name,
        )

    def _write_log_file(self, log_file: str, save_path: str):
        """Write audit logs to a file."""
        # Upload the log file to GCS
        blob = self.gcs_data.storage.get_blob(self.gcs_data.analysis_bucket, save_path)
        blob.upload_from_filename(log_file)
        return to_path(f'gs://{self.gcs_data.analysis_bucket}/{save_path}')

    def write_log_file(self, log_file: str, save_path: str) -> Path:
        """Write audit logs to a file."""
        return self._write_log_file(log_file, save_path)

    def write_reviewed_files_report(
        self,
        review_result: ReviewResult,
    ) -> Path:
        """
        Write reviewed files report to GCS.

        Args:
            review_result: The review analysis results
            bucket: GCS bucket for reports
            output_prefix: Optional prefix for output paths
        """
        return self._write_csv_report(
            blob_path=f'audit_results/{self.timestamp}/reviewed_files.csv',
            entries=review_result.reviewed_files,
            report_name='FILES REVIEWED',
        )

    def write_deleted_files_report(
        self, deletion_result: DeletionResult, dry_run: bool
    ) -> Path:
        """
        Write deleted files report to GCS.

        Args:
            deleted_files: The deleted files entries
            bucket: GCS bucket for reports
            dry_run: If True, will write to a dry run report
        """
        if dry_run:
            blob_path = f'audit_results/{self.timestamp}/deleted_files_dry_run.csv'
        else:
            blob_path = f'audit_results/{self.timestamp}/deleted_files.csv'
        return self._write_csv_report(
            blob_path=blob_path,
            entries=deletion_result.deleted_files,
            report_name='FILES DELETED',
        )

    def _write_unaligned_sgs_report(
        self,
        blob_path: str,
        unaligned_sgs: list[SequencingGroup],
    ):
        """Write report of sequencing groups without completed CRAMs."""
        if not unaligned_sgs:
            self.audit_logs.info_nl(
                'UNALIGNED SEQUENCING GROUPS - SKIPPED - No data to write'
            )
            return

        buffer = StringIO()
        writer = csv.writer(buffer, delimiter='\t')

        # Write header
        writer.writerow(
            [
                'SG ID',
                'SG Type',
                'SG Technology',
                'SG Platform',
                'Sample ID',
                'Sample External ID',
                'Participant ID',
                'Participant External ID',
                'Assay Count',
                'Total Read Size',
            ]
        )

        # Write data
        for sg in unaligned_sgs:
            sample_ext_id = sg.sample.external_id or ', '.join(
                sg.sample.external_ids.values()
            )
            participant_ext_id = sg.sample.participant.external_id or ', '.join(
                sg.sample.participant.external_ids.values()
            )

            writer.writerow(
                [
                    sg.id,
                    sg.type,
                    sg.technology,
                    sg.platform,
                    sg.sample.id,
                    sample_ext_id,
                    sg.sample.participant.id,
                    participant_ext_id,
                    len(sg.assays),
                    sg.get_total_read_size(),
                ]
            )
        blob = self.gcs_data.storage.get_blob(self.gcs_data.analysis_bucket, blob_path)
        self.gcs_data.storage.upload_from_buffer(
            blob, buffer, 'text/tab-separated-values'
        )
        self.audit_logs.info_nl(
            f'{"UNALIGNED SGS":<20}: gs://{self.gcs_data.analysis_bucket}/{blob_path} ({len(unaligned_sgs)} entries)'
        )

    # Summary statistics #
    def generate_summary_statistics(self, audit_result: AuditResult) -> dict[str, int]:
        """
        Generate summary statistics from audit results.

        Args:
            audit_result: The audit analysis results

        Returns:
            Dictionary of statistics
        """
        total_size_to_delete = sum(
            entry.filesize or 0 for entry in audit_result.files_to_delete
        )

        total_size_to_review = sum(
            entry.filesize or 0 for entry in audit_result.files_to_review
        )

        return {
            'files_to_delete': len(audit_result.files_to_delete),
            'files_to_delete_size_gb': total_size_to_delete / (1024**3),
            'files_to_review': len(audit_result.files_to_review),
            'files_to_review_size_gb': total_size_to_review / (1024**3),
            'unaligned_sgs': len(audit_result.unaligned_sequencing_groups),
        }

    # Filtering #
    def filter_rows(
        self, filter_expressions: list[str], rows: list[AuditReportEntry]
    ) -> list[AuditReportEntry]:
        """Filter report rows based on filter expressions."""
        if not filter_expressions:
            return rows

        filter_func = self.parse_filter_expressions(filter_expressions)

        filtered_rows = [
            row
            for row in rows
            if filter_func(row.to_report_dict(use_external_headers=False))
        ]
        self.audit_logs.info_nl(f'Filtered rows: {len(filtered_rows)} / {len(rows)}.')
        return filtered_rows

    def parse_filter_expressions(self, filter_expressions: list[str]) -> callable:
        """Parse filter expressions into a callable filter function."""

        def matches_filters(row: dict) -> bool:
            for expr in filter_expressions:
                if not evaluate_filter_expression(expr, row):
                    return False
            return True

        def evaluate_filter_expression(expr: str, row: dict) -> bool:
            """Evaluate a single filter expression against a row."""
            # Recursive cases first
            if ' or ' in expr:
                # Evaluate either side of the 'or'
                left, right = expr.split(' or ', 1)
                return evaluate_filter_expression(
                    left.strip(), row
                ) or evaluate_filter_expression(right.strip(), row)

            if ' and ' in expr:
                # Evaluate either side of the 'and'
                left, right = expr.split(' and ', 1)
                return evaluate_filter_expression(
                    left.strip(), row
                ) and evaluate_filter_expression(right.strip(), row)

            # Handle contains operator
            if ' contains ' in expr:
                field, value = expr.split(' contains ', 1)
                return value.strip('"\'') in str(row.get(field.strip(), ''))

            # Handle equality
            if '==' in expr:
                field, value = expr.split('==', 1)
                return str(row.get(field.strip(), '')).strip() == value.strip().strip(
                    '"\''
                )

            # Handle inequality
            if '!=' in expr:
                field, value = expr.split('!=', 1)
                return str(row.get(field.strip(), '')).strip() != value.strip().strip(
                    '"\''
                )
            if '>=' in expr:
                field, value = expr.split('>=', 1)
                return float(row.get(field.strip(), 0)) >= float(value.strip())
            if '<=' in expr:
                field, value = expr.split('<=', 1)
                return float(row.get(field.strip(), 0)) <= float(value.strip())
            if '>' in expr:
                field, value = expr.split('>', 1)
                return float(row.get(field.strip(), 0)) > float(value.strip())
            if '<' in expr:
                field, value = expr.split('<', 1)
                return float(row.get(field.strip(), 0)) < float(value.strip())

            return True

        return matches_filters



"""CLI entry point for deleting files in audit results."""

import click

from metamist.audit.data_access import GCSDataAccess, MetamistDataAccess
from metamist.audit.models import AuditReportEntry, DeletionResult
from metamist.audit.services import BucketAuditLogger, Reporter

from cpg_utils import Path, to_path


def delete_from_audit_results(
    dataset: str,
    results_folder: str,
    report: str,
    dry_run: bool = False,
):
    """Delete files listed in the audit results."""
    audit_logs = BucketAuditLogger(dataset, 'audit_deletions')

    gcs = GCSDataAccess(dataset)
    reporter = Reporter(gcs, audit_logs, results_folder)

    audit_logs.info_nl(f"Reading report '{report}'".center(50, '~'))

    report_rows = reporter.get_report_rows_from_name(report)

    deletion_result = delete_files_from_report(
        gcs, report, report_rows, audit_logs, dry_run
    )

    deletion_report = reporter.write_deleted_files_report(deletion_result, dry_run)
    stats = reporter.get_report_entries_stats(deletion_result.deleted_files)

    if not dry_run:
        upsert_deleted_files_analysis(
            dataset, report, results_folder, deletion_report, stats, audit_logs
        )

    all_stats = reporter.get_report_stats(deletion_report)

    audit_logs.info_nl(f"Finished processing deletes from report: '{report}'")
    audit_logs.info_nl(
        f'New deletions: {stats["file_count"]} files, {stats["total_size"] / (1024**3):.2f} GiB'
    )
    audit_logs.info_nl(
        f'Total deletions: {all_stats["file_count"]} files, {all_stats["total_size"] / (1024**3):.2f} GiB'
    )


# TODO - make sure we're tracking exactly which report the deletions came from
# as the reviewed_files report could be updated many times with different actions/comments
# and we want to track exactly which version of the report the deletions came from


# TODO - make this script agnostic of the report name, all actions and review comments should
# be handled in the review_audit_results.py script. This script should just delete files
# marked for deletion, and error if any files are not marked for deletion or missing comments.
# This will make it easier to maintain and reason about.
def delete_files_from_report(
    gcs: GCSDataAccess,
    report_name: str,
    rows: list[AuditReportEntry],
    audit_logs: BucketAuditLogger,
    dry_run: bool,
) -> DeletionResult:
    """
    Delete files in the upload bucket from the audit results.
    Only delete files from rows marked with 'DELETE' action.

    If deleting from 'files_to_delete', automatically add this action.
    """
    to_delete: list[AuditReportEntry] = []
    for row in rows:
        # TODO move this into the files_to_delete report generation
        if report_name == 'files_to_delete':
            row.update_action('DELETE')
            row.update_review_comment('Deleted due to audit determination')

        if not row.action or row.action.upper() != 'DELETE':
            continue

        if not row.review_comment:
            raise ValueError(
                "Review Comment is required for rows outside the 'files_to_delete' report. "
                'Use `review_audit_results.py` to add comments justifying the deletion, then try again.'
            )

        file_path = to_path(row.filepath)
        if not file_path.exists():
            audit_logs.warning(f'File {file_path} does not exist, skipping deletion.')
            continue

        to_delete.append(row)

    total_bytes = sum(int(row.filesize) for row in to_delete)
    if dry_run:
        audit_logs.info_nl(
            f'Dry run: would have deleted {len(to_delete)} files ({total_bytes / (1024**3):.2f} GiB)'
        )
    else:
        gcs.delete_blobs(
            gcs.upload_bucket,
            [to_path(row.filepath).blob for row in to_delete],
        )
        audit_logs.info_nl(
            f'Deleted: {len(to_delete)} files ({total_bytes / (1024**3):.2f} GiB)'
        )

    return DeletionResult(deleted_files=to_delete)


def upsert_deleted_files_analysis(
    dataset: str,
    audited_report_name: str,
    results_folder: str,
    deletion_report: Path,
    stats: dict,
    audit_logs: BucketAuditLogger,
):
    """
    Inserts or updates an analysis record for deleted files.
    Populates the meta dict with deleted files stats.
    """
    audit_logs.info_nl('Upserting analysis record for deleted files report.')
    if existing_analysis := MetamistDataAccess().get_audit_deletion_analysis(
        dataset, str(deletion_report)
    ):
        audit_logs.info_nl(
            f'Found existing analysis record (ID: {existing_analysis["id"]}) for {deletion_report}. Updating...'
        )
        MetamistDataAccess().update_audit_deletion_analysis(
            existing_analysis, audited_report_name, stats
        )
        audit_logs.info_nl(f'Updated analysis {existing_analysis["id"]}.')

    else:
        audit_logs.info_nl(f'Creating new record for {deletion_report}')
        cohort_name = f'{dataset}_{results_folder}'
        aid = MetamistDataAccess().create_audit_deletion_analysis(
            dataset, cohort_name, audited_report_name, str(deletion_report), stats
        )
        audit_logs.info_nl(f'Created analysis {aid}.')


@click.command()
@click.option('--dataset', '-d', required=True, help='Dataset name')
@click.option('--results-folder', '-r', required=True, help='Audit results folder')
@click.option(
    '--report-name',
    '-n',
    help="The report to delete files from. e.g. 'files_to_delete'.",
    default='files_to_delete',
)
@click.option('--dry-run', is_flag=True, help='If set, will not delete objects.')
def main(
    dataset: str,
    results_folder: str,
    report_name: str = 'files_to_delete',
    dry_run: bool = False,
):
    """
    Reads the report at location:
        gs://cpg-dataset-main-analysis/audit_results/{results_folder}/{report_name}.csv

    If the report is 'files_to_delete', it will delete the files listed in the report.

    For any other report, files must be marked for deletion in the report, alongside a
    comment justifying the deletion.

    Deleting files will create or update an analysis record of type 'audit_deletions'.
    """
    delete_from_audit_results(dataset, results_folder, report_name, dry_run)


if __name__ == '__main__':
    main()  # pylint: disable=no-value-for-parameter



import datetime
from random import randint

from db.python.filters import GenericFilter
from db.python.layers import CohortLayer, SampleLayer, SequencingGroupLayer
from db.python.tables.cohort import CohortFilter
from db.python.tables.sequencing_group import SequencingGroupFilter
from graphql.error import GraphQLError
from models.enums.cohort import CohortStatus, CohortUpdateStatus
from models.models.cohort import CohortCriteriaInternal, CohortUpdateBody
from models.utils.cohort_id_format import cohort_id_format
from models.utils.cohort_template_id_format import cohort_template_id_format
from test.testbase import DbIsolatedTest
from test.testbase import run_as_sync

from models.models import (
    PRIMARY_EXTERNAL_ORG,
    SampleUpsertInternal,
    SequencingGroupUpsertInternal,
)


def get_sample_model(eid, s_type='blood', sg_type='genome', plat='illumina'):
    """Create a minimal sample"""
    return SampleUpsertInternal(
        meta={},
        external_ids={PRIMARY_EXTERNAL_ORG: f'EXID{eid}'},
        type=s_type,
        sequencing_groups=[
            SequencingGroupUpsertInternal(
                type=sg_type,
                technology='short-read',
                platform=plat,
                meta={},
                assays=[],
            ),
            SequencingGroupUpsertInternal(
                type=sg_type,
                technology='long-read',
                platform=plat,
                meta={},
                assays=[],
            ),
        ],
    )


ACTIVE = 'active'
INVALID = 'invalid'
ARCHIVED = 'archived'


class TestStatusInCohortDBLayer(DbIsolatedTest):
    """Test cohort status related functions implemented in the DB layer"""

    @run_as_sync
    async def setUp(self):
        super().setUp()

        self.cohort_layer = CohortLayer(self.connection)
        self.sample_layer = SampleLayer(self.connection)

        self.sample_a = await self.sample_layer.upsert_sample(
            get_sample_model('A', 'saliva', 'exome', 'ONT')
        )
        self.sgA_raw = [sg.id for sg in self.sample_a.sequencing_groups]

        self.cohort_name = 'Sample cohort 1'
        self.cohort_description = 'Sample cohort description'
        self.cohort = await self.cohort_layer.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description=self.cohort_description,
            cohort_name=self.cohort_name,
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(
                projects=[self.project_id],
                sample_type=['saliva'],
            ),
        )

    @run_as_sync
    async def test_create_custom_cohort_and_verify_status(self):
        """Test to create a custom cohort and verify its status
        (Here the one created in the setup method is tested)
        """

        created_cohort_in_list = await self.cohort_layer.query(
            CohortFilter(id=GenericFilter(eq=self.cohort.cohort_id))
        )
        self.assertTrue(created_cohort_in_list)
        self.assertTrue(len(created_cohort_in_list) == 1)

        created_cohort = created_cohort_in_list[0]
        self.assertEqual(created_cohort.id, self.cohort.cohort_id)
        self.assertEqual(created_cohort.description, self.cohort_description)
        self.assertEqual(created_cohort.name, self.cohort_name)
        self.assertEqual(created_cohort.status, CohortStatus.active)

    @run_as_sync
    async def test_query_cohort_with_inactive_sample(self):
        """Test cohort status when inactive sample"""

        await self.sample_layer.upsert_sample(
            SampleUpsertInternal(id=self.sample_a.id, active=False)
        )
        cohort = (
            await self.cohort_layer.query(
                CohortFilter(id=GenericFilter(eq=self.cohort.cohort_id))
            )
        )[0]
        self.assertEqual(cohort.status, CohortStatus.invalid)

    @run_as_sync
    async def test_query_cohort_with_archived_sg(self):
        """Test cohort status when archived sequencing group"""

        await (SequencingGroupLayer(self.connection)).archive_sequencing_group(
            sequencing_group_id=self.sgA_raw[0]
        )
        cohort = (
            await self.cohort_layer.query(
                CohortFilter(id=GenericFilter(eq=self.cohort.cohort_id))
            )
        )[0]
        self.assertEqual(cohort.status, CohortStatus.invalid)

    @run_as_sync
    async def test_query_cohort_status_with_all_active(self):
        """Test computed cohort status when sample/s active,
        sg/s not archived and cohort status is active in the DB"""

        queried_sample = await self.sample_layer.get_by_id(sample_id=self.sample_a.id)
        self.assertTrue(queried_sample.active)

        queried_sg_list = await (SequencingGroupLayer(self.connection)).query(
            SequencingGroupFilter(
                id=GenericFilter(in_=[self.sgA_raw[0], self.sgA_raw[1]])
            )
        )
        self.assertFalse(queried_sg_list[0].archived)
        self.assertFalse(queried_sg_list[1].archived)

        # query directly from the cohort table as the returned status is computed runtime based on sample, sg and cohort
        cohort_raw_entry = await self.connection.connection.fetch_one(
            'SELECT status FROM cohort where id = :cohort_id',
            {'cohort_id': self.cohort.cohort_id},
        )
        self.assertEqual(dict(cohort_raw_entry)['status'], CohortStatus.active.value)

        cohort = (
            await self.cohort_layer.query(
                CohortFilter(id=GenericFilter(eq=self.cohort.cohort_id))
            )
        )[0]
        self.assertEqual(cohort.status, CohortStatus.active)

    @run_as_sync
    async def test_query_cohort_with_at_least_one_inactive_sample(self):
        """Test cohort status when at least one sample is inactive"""

        sample_b = await self.sample_layer.upsert_sample(get_sample_model('B'))
        sg_b_raw = [sg.id for sg in sample_b.sequencing_groups]

        new_cohort = self.cohort = await self.cohort_layer.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Sample cohort Test 2',
            cohort_name='Sample cohort 2',
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(
                projects=[self.project_id],
                sg_ids_internal_raw=[self.sgA_raw[0], sg_b_raw[0]],
            ),
        )
        await self.sample_layer.upsert_sample(
            SampleUpsertInternal(id=sample_b.id, active=False)
        )

        cohort = (
            await self.cohort_layer.query(
                CohortFilter(id=GenericFilter(eq=new_cohort.cohort_id))
            )
        )[0]
        self.assertEqual(cohort.status, CohortStatus.invalid)

    @run_as_sync
    async def test_query_cohort_with_archived_db_status(self):
        """Test computed cohort status when cohort is archived in the DB"""

        # directly update without using the cohort_db_layer
        await self.connection.connection.fetch_one(
            'UPDATE cohort SET status = :status WHERE id = :cohort_id',
            {
                'cohort_id': self.cohort.cohort_id,
                'status': CohortUpdateStatus.archived.value,
            },
        )

        cohort = (
            await self.cohort_layer.query(
                CohortFilter(id=GenericFilter(eq=self.cohort.cohort_id))
            )
        )[0]
        self.assertEqual(cohort.status, CohortStatus.archived)

    @run_as_sync
    async def test_query_cohort_in_get_template_by_cohort_id(self):
        """Test template query for retrieved based on cohort id"""
        template = await self.cohort_layer.get_template_by_cohort_id(
            self.cohort.cohort_id
        )
        self.assertTrue(template)


CREATE_COHORT_MUTATION = """
  mutation CreateCohortFromCriteria($project: String!, $cohortSpec: CohortBodyInput!, $cohortCriteria: CohortCriteriaInput!, $dryRun: Boolean) {
      cohort {
        createCohortFromCriteria(
          project: $project
          cohortSpec: $cohortSpec
          cohortCriteria: $cohortCriteria
          dryRun: $dryRun
        ) {
          status
        }
      }
    }
"""


class TestCohortStatusGraphQL(DbIsolatedTest):
    """Test cohort querying via GraphQL"""

    @run_as_sync
    async def setUp(self):
        super().setUp()

        self.cohort_layer = CohortLayer(self.connection)
        self.sample_a = await (SampleLayer(self.connection)).upsert_sample(
            get_sample_model('A', 'saliva', 'exome', 'ONT')
        )
        self.cohort_name = 'Sample cohort'
        self.cohort_description = 'Sample cohort description'

        self.cohort = await self.cohort_layer.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description=self.cohort_description,
            cohort_name=self.cohort_name,
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(
                projects=[self.project_id],
                sample_type=['saliva'],
            ),
        )
        self.cohort_id_formatted = cohort_id_format(self.cohort.cohort_id)

    @run_as_sync
    async def test_create_custom_cohort_response_for_status(self):
        """Test status field in create_custom_cohort mutation response"""

        mutation_result = (
            await self.run_graphql_query_async(
                CREATE_COHORT_MUTATION,
                variables={
                    'project': self.project_name,
                    'cohortSpec': {
                        'name': 'TestCohort1',
                        'description': 'TestCohortDescription',
                    },
                    'cohortCriteria': {
                        'projects': [self.project_name],
                        'sampleType': ['blood'],
                    },
                },
            )
        )['cohort']['createCohortFromCriteria']
        self.assertEqual(mutation_result['status'], ACTIVE)

    @run_as_sync
    async def test_query_cohort_with_filter_by_id(self):
        """Test status field in GraphQL query cohort (by id)"""

        query_cohort_incl_status = await self.run_graphql_query_async(
            """
            query CohortQuery($cohort_id: String!) {
                cohorts(id: {eq: $cohort_id}) {
                    name
                    status
                    description
                    sequencingGroups {
                        id
                        sample {
                            project {
                                name
                            }
                        }
                    }
                }
            }
        """,
            {'cohort_id': self.cohort_id_formatted},
        )

        self.assertEqual(len(query_cohort_incl_status['cohorts']), 1)
        queried_cohort = query_cohort_incl_status['cohorts'][0]

        self.assertEqual(queried_cohort['name'], self.cohort_name)
        self.assertEqual(queried_cohort['description'], self.cohort_description)

        self.assertEqual(
            queried_cohort['sequencingGroups'][0]['sample']['project']['name'],
            self.project_name,
        )
        self.assertEqual(
            queried_cohort['sequencingGroups'][1]['sample']['project']['name'],
            self.project_name,
        )
        self.assertEqual(queried_cohort['status'], ACTIVE)

    @run_as_sync
    async def test_query_cohort_with_filter_status_eq(self):
        """Test GraphQL query cohort with filter by status (eq)"""

        query_cohort_filter_status_eq = """
                    query CohortQuery($cohort_status: CohortStatus!) {
                cohorts(status: {eq: $cohort_status}) {
                    name
                    status
                }
            }
        """

        query_cohort_status_eq = await self.run_graphql_query_async(
            query_cohort_filter_status_eq,
            {'cohort_status': ACTIVE},
        )

        self.assertTrue(query_cohort_status_eq['cohorts'])
        queried_cohort = query_cohort_status_eq['cohorts'][0]

        self.assertEqual(queried_cohort['name'], self.cohort_name)
        self.assertEqual(queried_cohort['status'], ACTIVE)

        # update cohort status and retrieve
        await self.cohort_layer.update_cohort(
            CohortUpdateBody(status=CohortUpdateStatus.archived), self.cohort.cohort_id
        )
        query_cohort_status_eq = await self.run_graphql_query_async(
            query_cohort_filter_status_eq,
            {'cohort_status': ACTIVE},
        )
        self.assertFalse(query_cohort_status_eq['cohorts'])

    @run_as_sync
    async def test_query_cohort_with_filter_status_in(self):
        """Test GraphQL query cohort with filter by status (in)"""

        _ = (
            await self.run_graphql_query_async(
                CREATE_COHORT_MUTATION,
                variables={
                    'project': self.project_name,
                    'cohortSpec': {
                        'name': 'TestCohort1',
                        'description': 'TestCohortDescription',
                    },
                    'cohortCriteria': {
                        'projects': [self.project_name],
                        'sampleType': ['blood'],
                    },
                },
            )
        )['cohort']['createCohortFromCriteria']

        query_cohort_filter_status_in = """
            query CohortQuery($cohort_status_list: [CohortStatus!]!) {
                cohorts(status: {in_: $cohort_status_list}) {
                    status
                }
            }
        """

        query_cohort_status_in = await self.run_graphql_query_async(
            query_cohort_filter_status_in,
            {'cohort_status_list': [ACTIVE]},
        )
        self.assertTrue(len(query_cohort_status_in['cohorts']) == 2)
        for cohort in query_cohort_status_in['cohorts']:
            self.assertEqual(cohort['status'], ACTIVE)

        # update cohort status and retrieve
        await self.cohort_layer.update_cohort(
            CohortUpdateBody(status=CohortUpdateStatus.archived), self.cohort.cohort_id
        )

        query_cohort_status_in = await self.run_graphql_query_async(
            query_cohort_filter_status_in,
            {'cohort_status_list': [ACTIVE]},
        )
        self.assertTrue(len(query_cohort_status_in['cohorts']) == 1)

    @run_as_sync
    async def test_query_cohort_with_filter_status_nin(self):
        """Test GraphQL query cohort with filter by status (not in)"""

        query_cohort_status_nin = await self.run_graphql_query_async(
            """
            query CohortQuery($cohort_status_list: [CohortStatus!]!) {
                cohorts(status: {nin: $cohort_status_list}) {
                    status
                }
            }
        """,
            {'cohort_status_list': [ACTIVE]},
        )
        self.assertFalse(query_cohort_status_nin['cohorts'])

    @run_as_sync
    async def test_query_cohort_with_filter_status_criteria_not_defined(self):
        """Test GraphQL query cohort with filter by status (filter criteria is not one of eq, in or nin)"""

        query_cohorts = await self.run_graphql_query_async(
            """
            query CohortQuery($cohort_status: CohortStatus!) {
                cohorts(status: {gt: $cohort_status}) {
                    status
                }
            }
        """,
            {'cohort_status': ACTIVE},
        )
        self.assertTrue(query_cohorts['cohorts'])

    @run_as_sync
    async def test_query_info_of_non_existent_cohort(self):
        """Test GraphQL query cohort by non-existent cohort id"""

        query_cohorts = await self.run_graphql_query_async(
            """
            query CohortQuery($cohort_id: String!) {
                cohorts(id: {eq: $cohort_id}) {
                    status
                }
            }
        """,
            {'cohort_id': cohort_id_format(self.cohort.cohort_id + randint(1, 100))},
        )
        self.assertFalse(query_cohorts['cohorts'])

    @run_as_sync
    async def test_query_cohort_with_invalid_status_filter_value(self):
        """Test GraphQL query cohort by non-existent cohort status"""

        with self.assertRaises(GraphQLError):
            _ = await self.run_graphql_query_async(
                """
                query CohortQuery($cohort_status: CohortStatus!) {
                    cohorts(status: {eq: $cohort_status}) {
                    }
                }
            """,
                {'cohort_status': 'Dummy status'},
            )

    @run_as_sync
    async def test_update_cohort_fields(self):
        """Test GraphQL mutation for updating cohort fields"""

        new_name = 'Updated Llama'
        new_status = ARCHIVED
        new_description = 'Updated description'

        queried_cohort = (
            await self.run_graphql_query_async(
                """
            query CohortQuery($id: String!) {
                cohorts(id: {eq: $id}) {
                    status
                	name
                    id
                    description
                }
            }
        """,
                {'id': self.cohort_id_formatted},
            )
        )['cohorts'][0]

        self.assertNotEqual(queried_cohort['name'], new_name)
        self.assertNotEqual(queried_cohort['status'], ARCHIVED)
        self.assertNotEqual(queried_cohort['description'], new_description)

        updated_cohort = (
            await self.run_graphql_query_async(
                """
                mutation updateCohort($id : String!, $cohort: CohortUpdateBodyInput!)
                {
                  cohort{
                    updateCohort(id:$id, cohort:$cohort){
                      id
                      name
                      description
                      status
                    }
                  }

                }
        """,
                {
                    'id': self.cohort_id_formatted,
                    'cohort': {
                        'name': new_name,
                        'status': new_status,
                        'description': new_description,
                    },
                },
            )
        )['cohort']['updateCohort']

        self.assertEqual(updated_cohort['name'], new_name)
        self.assertEqual(updated_cohort['status'], ARCHIVED)
        self.assertEqual(updated_cohort['description'], new_description)

    @run_as_sync
    async def test_update_cohort_immutable_fields(self):
        """Test GraphQL mutation for updating cohort fields with immutable fields (not allowed to update)"""

        with self.assertRaises(GraphQLError):
            _ = await self.run_graphql_query_async(
                """
                    mutation updateCohort($id : String!, $cohort: CohortUpdateBodyInput!)
                    {
                      cohort{
                        updateCohort(id:$id, cohort:$cohort){
                          id
                        }
                      }
                    }
            """,
                {
                    'id': self.cohort_id_formatted,
                    'cohort': {
                        'author': 'Test author update',
                        'timestamp': datetime.datetime.now(),
                        'template_id': cohort_template_id_format(randint(1, 100)),
                        'project': self.project_id,
                    },
                },
            )

    @run_as_sync
    async def test_update_cohort_fields_with_empty_input_body(self):
        """Test GraphQL mutation for updating with empty body"""

        with self.assertRaises(GraphQLError):
            _ = await self.run_graphql_query_async(
                """
                    mutation updateCohort($id : String!, $cohort: CohortUpdateBodyInput!)
                    {
                      cohort{
                        updateCohort(id:$id, cohort:$cohort){
                          id
                        }
                      }
                    }
            """,
                {
                    'id': self.cohort_id_formatted,
                    'cohort': {},
                },
            )

    @run_as_sync
    async def test_update_non_existent_cohort(self):
        """Test GraphQL mutation for updating cohort fields of non-existent cohort"""

        with self.assertRaises(GraphQLError):
            _ = await self.run_graphql_query_async(
                """
                    mutation updateCohort($id : String!, $cohort: CohortUpdateBodyInput!)
                    {
                      cohort{
                        updateCohort(id:$id, cohort:$cohort){
                          id
                        }
                      }
                    }
            """,
                {
                    'id': cohort_id_format(self.cohort.cohort_id + randint(1, 100)),
                    'cohort': {'name': 'Test name change'},
                },
            )

    @run_as_sync
    async def test_update_status_of_invalid_cohort(self):
        """Test GraphQL mutation for updating status of an INVALID cohort"""

        await SampleLayer(self.connection).upsert_sample(
            SampleUpsertInternal(id=self.sample_a.id, active=False)
        )
        cohort = (
            await self.cohort_layer.query(
                CohortFilter(id=GenericFilter(eq=self.cohort.cohort_id))
            )
        )[0]
        self.assertEqual(cohort.status, CohortStatus.invalid)

        updated_cohort = (
            await self.run_graphql_query_async(
                """
                    mutation updateCohort($id : String!, $cohort: CohortUpdateBodyInput!)
                    {
                      cohort{
                        updateCohort(id:$id, cohort:$cohort){
                          status
                        }
                      }
                    }
            """,
                {
                    'id': cohort_id_format(self.cohort.cohort_id),
                    'cohort': {'status': ACTIVE},
                },
            )
        )['cohort']['updateCohort']

        self.assertEqual(updated_cohort['status'], INVALID)

    @run_as_sync
    async def test_update_status_of_archived_cohort_with_archived_samples(self):
        """Test GraphQL mutation for updating status of an archived cohort with archived samples"""

        # directly update cohort DB status
        await self.connection.connection.fetch_one(
            'UPDATE cohort SET status = :status WHERE id = :cohort_id',
            {
                'cohort_id': self.cohort.cohort_id,
                'status': CohortUpdateStatus.archived.value,
            },
        )
        await SampleLayer(self.connection).upsert_sample(
            SampleUpsertInternal(id=self.sample_a.id, active=False)
        )
        cohort = (
            await self.cohort_layer.query(
                CohortFilter(id=GenericFilter(eq=self.cohort.cohort_id))
            )
        )[0]
        self.assertEqual(cohort.status, CohortStatus.archived)

        with self.assertRaises(GraphQLError):
            _ = await self.run_graphql_query_async(
                """
                    mutation updateCohort($id : String!, $cohort: CohortUpdateBodyInput!)
                    {
                      cohort{
                        updateCohort(id:$id, cohort:$cohort){
                          id
                        }
                      }
                    }
            """,
                {
                    'id': cohort_id_format(self.cohort.cohort_id),
                    'cohort': {'status': ACTIVE},
                },
            )

    @run_as_sync
    async def test_update_status_of_archived_cohort_with_active_samples(self):
        """Test GraphQL mutation for updating status of an archived cohort with archived samples"""

        # directly update cohort DB status
        await self.connection.connection.fetch_one(
            'UPDATE cohort SET status = :status WHERE id = :cohort_id',
            {
                'cohort_id': self.cohort.cohort_id,
                'status': CohortUpdateStatus.archived,
            },
        )

        updated_cohort = (
            await self.run_graphql_query_async(
                """
                    mutation updateCohort($id : String!, $cohort: CohortUpdateBodyInput!)
                    {
                      cohort{
                        updateCohort(id:$id, cohort:$cohort){
                          status
                        }
                      }
                    }
            """,
                {
                    'id': cohort_id_format(self.cohort.cohort_id),
                    'cohort': {'status': ACTIVE},
                },
            )
        )['cohort']['updateCohort']

        self.assertEqual(updated_cohort['status'], ACTIVE)
